# -*- coding: utf-8 -*-

# Copyright (C) 2026 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Paths to files and directories used."""

import sys

from pathlib import Path

def get_data_directory():
    """Return path to the gaupol data directory."""
    # Data files are shipped inside the Python package,
    # which works the same in all contexts.
    return Path(__file__).resolve().parent / "data"

def get_locale_directory():
    """Return path to the locale directory."""
    if hasattr(sys, "frozen"):
        # Windows bundled exe
        prefix = Path(sys.executable).parent.parent
        return (prefix / "share" / "locale").resolve()
    # Running from the source directory.
    root = Path(__file__).resolve().parent.parent
    return (root / "locale").resolve()

DATA_DIR = get_data_directory()

# Overwritten when installing.
LOCALE_DIR = Path('/usr/share/locale')



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Dialog for displaying a traceback in case of an unhandled exception."""

import aeidon
import gaupol
import os
import platform
import sys
import traceback

from aeidon.i18n   import _
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

class DebugDialog(Gtk.MessageDialog):

    """Dialog for displaying a traceback in case of an unhandled exception."""

    def __init__(self, parent=None):
        """Initialize a :class:`DebugDialog` instance."""
        GObject.GObject.__init__(self,
                                 message_type=Gtk.MessageType.ERROR,
                                 transient_for=parent,
                                 text=_("Something went wrong"),
                                 secondary_text=_("You have probably discovered a bug. Please report it by providing the below information and a description of what you were doing."))

        self._text_view = Gtk.TextView()
        self._init_dialog()
        self._init_text_view()

    def _init_dialog(self):
        """Initialize the dialog."""
        self.add_button(_("_Report Bug"), Gtk.ResponseType.HELP)
        self.add_button(_("_Quit"), Gtk.ResponseType.NO)
        self.add_button(_("_Close"), Gtk.ResponseType.CLOSE)
        self.set_default_response(Gtk.ResponseType.CLOSE)
        self.set_modal(True)
        aeidon.util.connect(self, self, "response")

    def _init_text_view(self):
        """Initialize the text view."""
        self._text_view.set_wrap_mode(Gtk.WrapMode.WORD)
        self._text_view.set_editable(False)
        self._text_view.set_cursor_visible(False)
        self._text_view.set_accepts_tab(False)
        self._text_view.set_left_margin(6)
        self._text_view.set_right_margin(6)
        self._text_view.set_top_margin(6)
        self._text_view.set_bottom_margin(6)
        text_buffer = self._text_view.get_buffer()
        text_buffer.create_tag("monospace", family="monospace")
        text_buffer.create_tag("bold", weight=Pango.Weight.BOLD)
        text_buffer.create_tag("large", scale=1.1)
        scroller = Gtk.ScrolledWindow()
        scroller.set_policy(*((Gtk.PolicyType.AUTOMATIC,)*2))
        scroller.set_has_frame(True)
        scroller.set_child(self._text_view)
        box = self.get_message_area()
        gaupol.util.pack_start_expand(box, scroller)

    def _insert_environment(self):
        """Insert information about user's platform and environment."""
        locale = aeidon.locales.get_system_code()
        encoding = aeidon.encodings.get_locale_code()
        wayland = "yes" if (
            os.getenv("XDG_SESSION_TYPE", "") == "wayland" or
            os.getenv("WAYLAND_DISPLAY", "")
        ) else "no"
        self._insert_text(f"Platform: {platform.platform(True)}\n")
        self._insert_text(f"Locale: {locale}.{encoding}\n")
        self._insert_text(f"Wayland: {wayland}\n\n")

    def _insert_dependencies(self):
        """Insert version numbers of selected dependencies."""
        dotjoin = lambda x: ".".join(map(str, x))
        chardet_version = aeidon.util.get_chardet_version()
        gst_version = gaupol.util.get_gst_version()
        libspelling_version = gaupol.util.get_libspelling_version()
        gtk_version = dotjoin((
            Gtk.get_major_version(),
            Gtk.get_minor_version(),
            Gtk.get_micro_version(),
        ))
        pygobject_version = dotjoin(GObject.pygobject_version)
        python_version = dotjoin(sys.version_info[:3])
        self._insert_text(f"aeidon: {aeidon.__version__}\n")
        self._insert_text(f"charset-normalizer: {chardet_version}\n")
        self._insert_text(f"gaupol: {gaupol.__version__}\n")
        self._insert_text(f"gstreamer: {gst_version}\n")
        self._insert_text(f"gtk+: {gtk_version}\n")
        self._insert_text(f"libspelling: {libspelling_version}\n")
        self._insert_text(f"pygobject: {pygobject_version}\n")
        self._insert_text(f"python: {python_version}\n")

    def _insert_text(self, text, *tags):
        """Insert `text` with `tags` to the text view."""
        text_buffer = self._text_view.get_buffer()
        itr = text_buffer.get_end_iter()
        tags = tags + ("monospace",)
        text_buffer.insert_with_tags_by_name(itr, text, *tags)

    def _insert_title(self, text):
        """Insert `text` as a title to the text view."""
        self._insert_text(text, "large", "bold")
        self._insert_text("\n\n")

    def _insert_traceback(self, exc, tb, limit=100):
        """Insert up to `limit` stack trace entries from `tb`."""
        for line in traceback.format_tb(tb, limit=limit):
            self._insert_text(line)
        self._insert_text("\n")
        for line in traceback.format_exception_only(exc):
            self._insert_text(line)
        self._insert_text("\n")

    def _on_response(self, dialog, response):
        """Do not send response if reporting bug."""
        if response != Gtk.ResponseType.HELP: return
        gaupol.util.show_uri(gaupol.BUG_REPORT_URL)
        self.stop_emission("response")

    def set_text(self, exctype, exc, tb):
        """Set text from `tb` to the text view."""
        self._insert_title("Traceback")
        self._insert_traceback(exc, tb)
        self._insert_title("Environment")
        self._insert_environment()
        self._insert_title("Versions")
        self._insert_dependencies()
        gaupol.util.scale_to_content(self._text_view,
                                     min_nchar=60,
                                     max_nchar=100,
                                     min_nlines=10,
                                     max_nlines=25,
                                     font="monospace")



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""GTK user interface for the Gaupol subtitle editor."""

__version__ = "2.0"
COMBO_SEPARATOR = "<separator/>"

import contextlib
import sys
import warnings

if hasattr(sys, "frozen"):
    # Avoid error trying to write to non-existent stderr.
    # https://stackoverflow.com/a/35773092
    warnings.simplefilter("ignore")

# Needed for the spell-check dialog, where the '_proceed' function call is
# recursive. The recursion limit effectively defines the maximum allowed
# amount of consecutive subtitles without spelling errors, which if exceeded
# triggers a RecursionError.
sys.setrecursionlimit(10000)

import aeidon
import gi
import os

# Disable gst-vaapi as it doesn't seem to work with the GTK video sink.
# https://github.com/otsaloma/gaupol/issues/79
os.environ["LIBVA_DRIVER_NAME"] = "null"
os.environ["LIBVA_DRIVERS_PATH"] = "/dev/null"

gi.require_version("Gdk", "4.0")
gi.require_version("Graphene", "1.0")
gi.require_version("Gsk", "4.0")
gi.require_version("Gtk", "4.0")

with contextlib.suppress(Exception):
    gi.require_version("GtkSource", "5")

for module, version in {
    "Gst": "1.0",
    "GstPbutils": "1.0",
    "GstVideo": "1.0",
    "GstTag": "1.0",
}.items():
    with contextlib.suppress(Exception):
        gi.require_version(module, version)

with contextlib.suppress(Exception):
    from gi.repository import Gst
    Gst.init(None)

from gaupol.paths import DATA_DIR
from gaupol.paths import LOCALE_DIR
from gaupol.urls import BUG_REPORT_URL
from gaupol.urls import DOCUMENTATION_URL
from gaupol.urls import HOMEPAGE_URL
from gaupol.urls import REGEX_HELP_URL
from gaupol import util
from gaupol.enums import fields
from gaupol.enums import length_units
from gaupol.enums import orientation
from gaupol.enums import targets
from gaupol.enums import themes
from gaupol.errors import Default
from gaupol.attrdict import AttributeDictionary
from gaupol.config import ConfigurationStore
conf = ConfigurationStore()
from gaupol import style
from gaupol import ruler
from gaupol.entries import TimeEntry
from gaupol.renderers import FloatCellRenderer
from gaupol.renderers import IntegerCellRenderer
from gaupol.renderers import MultilineCellRenderer
from gaupol.renderers import MultilineDiffCellRenderer
from gaupol.renderers import TimeCellRenderer
from gaupol.toast import Toast
from gaupol.spell import SpellChecker
from gaupol.view import View
from gaupol.page import Page
from gaupol.player import VideoPlayer
from gaupol.dialogs.builder import BuilderDialog
from gaupol.dialogs.file import FileDialog
from gaupol.dialogs.open import OpenDialog
from gaupol.dialogs.save import SaveDialog
from gaupol.dialogs.video import VideoDialog
from gaupol.dialogs.append import AppendDialog
from gaupol.dialogs.about import AboutDialog
from gaupol.dialogs.debug import DebugDialog
from gaupol.dialogs.duration_adjust import DurationAdjustDialog
from gaupol.dialogs.encoding import EncodingDialog
from gaupol.dialogs.encoding import MenuEncodingDialog
from gaupol.dialogs.framerate_convert import FramerateConvertDialog
from gaupol.dialogs.insert import InsertDialog
from gaupol.dialogs.language import LanguageDialog
from gaupol.dialogs.message import ErrorDialog
from gaupol.dialogs.message import InfoDialog
from gaupol.dialogs.message import QuestionDialog
from gaupol.dialogs.message import WarningDialog
from gaupol.dialogs.multi_close import MultiCloseDialog
from gaupol.dialogs.multi_save import MultiSaveDialog
from gaupol.dialogs.preferences import PreferencesDialog
from gaupol.dialogs.preview_error import PreviewErrorDialog
from gaupol.dialogs.search import SearchDialog
from gaupol.dialogs.position_shift import FrameShiftDialog
from gaupol.dialogs.position_shift import TimeShiftDialog
from gaupol.dialogs.spell_check import SpellCheckDialog
from gaupol.dialogs.split import SplitDialog
from gaupol.dialogs.position_transform import FrameTransformDialog
from gaupol.dialogs.position_transform import TimeTransformDialog
from gaupol.assistants import TextAssistant
from gaupol.assistants import TextAssistantPage
from gaupol.action import Action
from gaupol.action import OpenRecentTranslationFileAction
from gaupol.action import RadioAction
from gaupol.action import ToggleAction
from gaupol import actions
from gaupol import agents
from gaupol.application import Application
from gaupol.applicationman import ApplicationManager
from gaupol.unittest import TestCase

def main(args):
    """Initialize application."""
    global appman
    # aeidon can be installed separately from gaupol, e.g. from a distro
    # package or PyPI, in which case the versions can be out of sync.
    if aeidon.__version__ != __version__:
        print("\n".join((
            f"Warning: aeidon and gaupol version mismatch:", # noqa
            f"{aeidon.__version__} {aeidon.__file__}",
            f"{__version__} {__file__}",
        )), file=sys.stderr)
    aeidon.i18n.bind(LOCALE_DIR)
    appman = ApplicationManager(args)
    raise SystemExit(appman.run())



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Base class for GTK unit test cases."""

import aeidon
import gaupol
import time

from gi.repository import GLib

class TestCase(aeidon.TestCase):

    """
    Base class for GTK unit test cases.

    Unit tests are designed to be run with ``py.test``, ``nose`` or something
    compatible. Tests should use plain ``assert`` statements to allow multiple
    different tools to be used to run the tests.
    """

    def main_loop(self, window):
        """Iterate the main loop until `window` is closed."""
        context = GLib.MainContext.default()
        while window.get_visible():
            while context.pending():
                context.iteration(False)
            time.sleep(0.01)

    def new_application(self):
        """Return a new application with one open page."""
        application = gaupol.Application()
        application.add_page(self.new_page())
        application.window.present()
        return application

    def new_page(self):
        """Return a new page with two open documents."""
        page = gaupol.Page()
        page.project.open_main(self.new_subrip_file(), "ascii")
        page.project.open_translation(self.new_microdvd_file(), "ascii")
        return page

    def teardown_method(self, method):
        """Remove state set for executing tests in `method`."""
        gaupol.util.iterate_main()
        for name in ("assistant", "dialog", "window"):
            if hasattr(self, name):
                getattr(self, name).destroy()
        if hasattr(self, "application"):
            self.application.window.set_visible(False)
            self.application.window.destroy()
        gaupol.util.iterate_main()
        gaupol.conf.restore_defaults()



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Dialogs for selecting character encodings."""

import aeidon
import gaupol

from aeidon.i18n   import _
from gi.repository import GObject
from gi.repository import Gtk

class EncodingDialog(Gtk.Dialog):

    """Dialog for selecting a character encoding."""

    def __init__(self, parent):
        """Initialize an :class:`EncodingDialog` instance."""
        GObject.GObject.__init__(self, use_header_bar=True)
        self._encoding = None
        self._encoding_list = Gtk.ListBox()
        self._init_dialog(parent)
        self._init_encoding_list()

    def get_encoding(self):
        """Return the selected encoding or ``None``."""
        return self._encoding

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_OK"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_modal(True)
        self.set_title(_("Character Encodings"))

    def _init_encoding_list(self):
        """Initialize the character encoding list."""
        self._encoding_list.set_selection_mode(Gtk.SelectionMode.NONE)
        self._encoding_list.set_show_separators(True)
        self._encoding_list.add_css_class("rich-list")
        self._encoding_list.connect("row-activated", self._on_encoding_list_row_activated)
        group = None
        for code, name, description in sorted(
                aeidon.encodings.get_valid(), key=lambda x: x[2]):
            box = gaupol.util.new_hbox(spacing=12)
            label = Gtk.Label(label=description)
            label.add_css_class("dim-label")
            box.append(label)
            box.append(Gtk.Label(label=name))
            # Arrow keys in a list box move focus between rows, not between
            # radio buttons, so keep the radio buttons out of the focus chain
            # and let the rows handle keyboard focus and activation.
            radio = Gtk.CheckButton(hexpand=True, can_focus=False, group=group)
            radio.set_child(box)
            radio.gaupol_code = code
            radio.connect("toggled", self._on_encoding_radio_toggled)
            group = group or radio
            row = Gtk.ListBoxRow(activatable=True, selectable=False)
            row.set_child(self._create_row_child(radio))
            row.gaupol_radio = radio
            self._encoding_list.append(row)
        frame = Gtk.Frame(child=self._encoding_list)
        content_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        gaupol.util.set_widget_margins(content_box, 18)
        content_box.append(frame)
        scroller = Gtk.ScrolledWindow(
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            propagate_natural_width=True,
            propagate_natural_height=True)
        scroller.set_max_content_height(gaupol.util.lines_to_px(25))
        scroller.set_child(content_box)
        gaupol.util.pack_start_expand(self.get_content_area(), scroller)

    def _create_row_child(self, radio):
        """Return the widget to place in `radio`'s list box row."""
        return radio

    def _on_encoding_list_row_activated(self, list_box, row):
        """Select the encoding of the activated row."""
        row.gaupol_radio.set_active(True)

    def _on_encoding_radio_toggled(self, radio):
        """Save the selected encoding."""
        if radio.get_active():
            self._encoding = radio.gaupol_code

class MenuEncodingDialog(EncodingDialog):

    """Dialog for selecting character encodings."""

    @aeidon.deco.listify
    def get_visible_encodings(self):
        """Return encodings chosen to be visible."""
        for i in range(1_000_000):
            row = self._encoding_list.get_row_at_index(i)
            if not row: break
            toggle = row.get_child().get_last_child()
            if toggle.get_active():
                yield toggle.gaupol_code

    def _create_row_child(self, radio):
        """Return a row widget with an "In Menu" toggle added."""
        toggle = Gtk.ToggleButton(label=_("In Menu"))
        toggle.set_valign(Gtk.Align.CENTER)
        toggle.set_active(radio.gaupol_code in gaupol.conf.encoding.visible)
        toggle.gaupol_code = radio.gaupol_code
        box = Gtk.Box()
        box.append(radio)
        box.append(toggle)
        return box



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Dialog for editing preferences."""

import aeidon
import gaupol

from aeidon.i18n   import _
from gi.repository import Gtk
from gi.repository import Pango

class EditorPage(aeidon.Delegate, gaupol.BuilderDialog):

    """Editor preferences page."""

    _widgets = [
        "editor_custom_font_check",
        "editor_font_button",
        "editor_length_cell_check",
        "editor_length_combo",
        "editor_length_edit_check",
        "editor_length_hbox",
        "editor_spell_check_check",
    ]

    def __init__(self, master, application):
        """Initialize an :class:`EditorPage` instance."""
        aeidon.Delegate.__init__(self, master)
        self._set_attributes(self._widgets, "editor_")
        self.application = application
        self._init_length_combo()
        self._init_values()

    def _get_custom_font(self):
        """Return custom font as string."""
        context = Gtk.Label().get_pango_context()
        font_desc = context.get_font_description()
        font = gaupol.conf.editor.custom_font
        custom_font_desc = Pango.FontDescription(font)
        font_desc.merge(custom_font_desc, True)
        return font_desc.to_string()

    def _init_length_combo(self):
        """Initialize the line length combo box."""
        store = Gtk.ListStore(str)
        for label in (x.label for x in gaupol.length_units):
            store.append((label,))
        self._length_combo.set_model(store)
        renderer = Gtk.CellRendererText()
        self._length_combo.pack_start(renderer, expand=True)
        self._length_combo.add_attribute(renderer, "text", 0)

    def _init_values(self):
        """Initialize default values for widgets."""
        use_custom = gaupol.conf.editor.use_custom_font
        self._custom_font_check.set_active(use_custom)
        self._font_button.set_sensitive(use_custom)
        self._font_button.set_font(self._get_custom_font())
        show_cell = gaupol.conf.editor.show_lengths_cell
        show_edit = gaupol.conf.editor.show_lengths_edit
        self._length_hbox.set_sensitive(show_cell or show_edit)
        self._length_cell_check.set_active(show_cell)
        self._length_edit_check.set_active(show_edit)
        self._length_combo.set_active(gaupol.conf.editor.length_unit)
        if gaupol.SpellChecker.available():
            inline = gaupol.conf.spell_check.inline
            self._spell_check_check.set_active(inline)
            self._spell_check_check.set_sensitive(True)
        else: # not available
            self._spell_check_check.set_active(False)
            self._spell_check_check.set_sensitive(False)

    def _on_custom_font_check_toggled(self, check_button):
        """Save custom font usage."""
        use_custom = check_button.get_active()
        gaupol.conf.editor.use_custom_font = use_custom
        self._font_button.set_sensitive(use_custom)

    def _on_font_button_font_set(self, font_button):
        """Save custom font."""
        gaupol.conf.editor.custom_font = font_button.get_font()

    def _on_length_cell_check_toggled(self, check_button):
        """Save line length display on cells."""
        gaupol.conf.editor.show_lengths_cell = check_button.get_active()
        show_cell = gaupol.conf.editor.show_lengths_cell
        show_edit = gaupol.conf.editor.show_lengths_edit
        self._length_hbox.set_sensitive(show_cell or show_edit)

    def _on_length_combo_changed(self, combo_box):
        """Save line length unit."""
        index = combo_box.get_active()
        gaupol.conf.editor.length_unit = gaupol.length_units[index]

    def _on_length_edit_check_toggled(self, check_button):
        """Save line length display on text views."""
        gaupol.conf.editor.show_lengths_edit = check_button.get_active()
        show_cell = gaupol.conf.editor.show_lengths_cell
        show_edit = gaupol.conf.editor.show_lengths_edit
        self._length_hbox.set_sensitive(show_cell or show_edit)

    def _on_spell_check_check_toggled(self, check_button):
        """Save inline spell-check use on text views."""
        gaupol.conf.spell_check.inline = check_button.get_active()

class FilePage(aeidon.Delegate, gaupol.BuilderDialog, gaupol.FileDialog):

    """File preferences page."""

    _widgets = [
        "file_add_button",
        "file_auto_check",
        "file_down_button",
        "file_encoding_list",
        "file_encoding_scroller",
        "file_force_encoding_check",
        "file_force_encoding_combo",
        "file_force_newline_check",
        "file_force_newline_combo",
        "file_locale_check",
        "file_remove_button",
        "file_up_button",
    ]

    def __init__(self, master, application):
        """Initialize a :class:`FilePage` instance."""
        aeidon.Delegate.__init__(self, master)
        self._set_attributes(self._widgets, "file_")
        # The encoding combo handling inherited from FileDialog,
        # which populates and reads it, operates on _encoding_combo.
        self._encoding_combo = self._force_encoding_combo
        self.application = application
        max_height = gaupol.util.lines_to_px(15)
        self._encoding_scroller.set_max_content_height(max_height)
        self._init_encoding_combo()
        self._init_force_newline_combo()
        self._init_values()

    def _get_selected_row(self):
        """Return the index of the selected row or ``None``."""
        row = self._encoding_list.get_selected_row()
        return row.get_index() if row else None

    def _init_force_newline_combo(self):
        """Initialize the newline combo box."""
        store = Gtk.ListStore(str)
        self._force_newline_combo.set_model(store)
        for name in (x.label for x in aeidon.newlines):
            store.append((name,))
        view = self._force_newline_combo.get_child()
        view.set_displayed_row(gaupol.util.tree_row_to_path(0))
        renderer = Gtk.CellRendererText()
        self._force_newline_combo.pack_start(renderer, expand=True)
        self._force_newline_combo.add_attribute(renderer, "text", 0)

    def _init_values(self):
        """Initialize default values for widgets."""
        self._auto_check.set_active(gaupol.conf.encoding.try_auto)
        self._auto_check.set_sensitive(aeidon.util.chardet_available())
        self._locale_check.set_active(gaupol.conf.encoding.try_locale)
        self._reload_encoding_list()
        encoding = gaupol.conf.file.force_encoding
        self._force_encoding_check.set_active(bool(encoding))
        self._force_encoding_combo.set_sensitive(bool(encoding))
        self.set_encoding(encoding)
        newline = gaupol.conf.file.force_newline
        self._force_newline_check.set_active(bool(newline))
        self._force_newline_combo.set_sensitive(bool(newline))
        self._force_newline_combo.set_active(
            newline or aeidon.util.get_default_newline())

    def _on_add_button_clicked(self, *args):
        """Add a new fallback encoding."""
        dialog = gaupol.EncodingDialog(self._dialog)
        response = gaupol.util.run_dialog(dialog)
        encoding = dialog.get_encoding()
        dialog.destroy()
        if response != Gtk.ResponseType.OK: return
        if encoding is None: return
        if encoding in gaupol.conf.encoding.fallback: return
        gaupol.conf.encoding.fallback.append(encoding)
        self._reload_encoding_list()
        self._select_row(len(gaupol.conf.encoding.fallback) - 1)

    def _on_auto_check_toggled(self, check_button):
        """Save encoding auto-detection usage."""
        gaupol.conf.encoding.try_auto = check_button.get_active()

    def _on_down_button_clicked(self, *args):
        """Move the selected fallback encoding down."""
        row = self._get_selected_row()
        encodings = gaupol.conf.encoding.fallback
        encodings.insert(row + 1, encodings.pop(row))
        self._reload_encoding_list()
        self._select_row(row + 1)

    def _on_encoding_list_row_selected(self, list_box, row):
        """Set the fallback list button sensitivities."""
        self._set_sensitivities()

    def _on_force_encoding_check_toggled(self, check_button):
        """Save the encoding to always convert to when saving."""
        force = check_button.get_active()
        self._force_encoding_combo.set_sensitive(force)
        gaupol.conf.file.force_encoding = (
            self.get_encoding() or "" if force else "")

    def _on_force_encoding_combo_changed(self, *args):
        """Save the encoding to always convert to when saving."""
        gaupol.FileDialog._on_encoding_combo_changed(self, *args)
        encoding = self.get_encoding()
        if encoding in (None, "other"): return
        gaupol.conf.file.force_encoding = encoding

    def _on_force_newline_check_toggled(self, check_button):
        """Save the newlines to always convert to when saving."""
        force = check_button.get_active()
        self._force_newline_combo.set_sensitive(force)
        gaupol.conf.file.force_newline = (
            aeidon.newlines[self._force_newline_combo.get_active()]
            if force else None)

    def _on_force_newline_combo_changed(self, combo_box):
        """Save the newlines to always convert to when saving."""
        gaupol.conf.file.force_newline = aeidon.newlines[combo_box.get_active()]

    def _on_locale_check_toggled(self, check_button):
        """Save locale encoding usage."""
        gaupol.conf.encoding.try_locale = check_button.get_active()

    def _on_remove_button_clicked(self, *args):
        """Remove the selected fallback encoding."""
        row = self._get_selected_row()
        gaupol.conf.encoding.fallback.pop(row)
        self._reload_encoding_list()
        if not gaupol.conf.encoding.fallback: return
        self._select_row(max(row - 1, 0))

    def _on_up_button_clicked(self, *args):
        """Move the selected fallback encoding up."""
        row = self._get_selected_row()
        encodings = gaupol.conf.encoding.fallback
        encodings.insert(row - 1, encodings.pop(row))
        self._reload_encoding_list()
        self._select_row(row - 1)

    def _reload_encoding_list(self):
        """Reload the fallback encoding list."""
        self._encoding_list.remove_all()
        for encoding in gaupol.conf.encoding.fallback:
            box = gaupol.util.new_hbox(spacing=12)
            label = Gtk.Label(
                label=aeidon.encodings.code_to_description(encoding))
            label.add_css_class("dim-label")
            box.append(label)
            box.append(Gtk.Label(
                label=aeidon.encodings.code_to_name(encoding)))
            self._encoding_list.append(box)
        self._set_sensitivities()

    def _select_row(self, index):
        """Select and focus the fallback encoding row at `index`."""
        row = self._encoding_list.get_row_at_index(index)
        self._encoding_list.select_row(row)
        row.grab_focus()

    def _set_sensitivities(self):
        """Set the fallback list button sensitivities."""
        row = self._get_selected_row()
        row = -1 if row is None else row
        count = len(gaupol.conf.encoding.fallback)
        self._remove_button.set_sensitive(row >= 0)
        self._up_button.set_sensitive(row > 0)
        self._down_button.set_sensitive(0 <= row < count - 1)

class PreviewPage(aeidon.Delegate, gaupol.BuilderDialog):

    """Preview preferences page."""

    _widgets = [
        "preview_app_combo",
        "preview_command_entry",
        "preview_force_utf_8_check",
        "preview_offset_spin",
    ]

    def __init__(self, master, application):
        """Initialize a :class:`PreviewPage` instance."""
        aeidon.Delegate.__init__(self, master)
        self._set_attributes(self._widgets, "preview_")
        self.application = application
        self._init_app_combo()
        self._init_values()

    def _init_app_combo(self):
        """Initialize the application combo box."""
        store = Gtk.ListStore(str)
        self._app_combo.set_model(store)
        for label in (x.label for x in aeidon.players):
            store.append((label,))
        store.append((gaupol.COMBO_SEPARATOR,))
        store.append((_("Custom"),))
        renderer = Gtk.CellRendererText()
        self._app_combo.pack_start(renderer, expand=True)
        self._app_combo.add_attribute(renderer, "text", 0)
        func = gaupol.util.separate_combo
        self._app_combo.set_row_separator_func(func, None)

    def _init_values(self):
        """Initialize default values for widgets."""
        store = self._app_combo.get_model()
        self._app_combo.set_active((len(store) - 1 if
                                    gaupol.conf.preview.use_custom_command else
                                    gaupol.conf.preview.player))

        self._command_entry.set_text(gaupol.util.get_preview_command())
        editable = gaupol.conf.preview.use_custom_command
        self._command_entry.set_editable(editable)
        self._force_utf_8_check.set_active(gaupol.conf.preview.force_utf_8)
        self._offset_spin.set_value(gaupol.conf.preview.offset)

    def _on_app_combo_changed(self, combo_box):
        """Save video player and show it's command."""
        index = combo_box.get_active()
        use_custom = not index in aeidon.players
        gaupol.conf.preview.use_custom_command = use_custom
        if not use_custom:
            gaupol.conf.preview.player = aeidon.players[index]
        self._command_entry.set_text(gaupol.util.get_preview_command())
        editable = gaupol.conf.preview.use_custom_command
        self._command_entry.set_editable(editable)

    def _on_command_entry_changed(self, entry):
        """Save custom command."""
        if not gaupol.conf.preview.use_custom_command: return
        gaupol.conf.preview.custom_command = entry.get_text()

    def _on_force_utf_8_check_toggled(self, check_button):
        """Save forced UTF-8 preview setting."""
        gaupol.conf.preview.force_utf_8 = check_button.get_active()
        self._command_entry.set_text(gaupol.util.get_preview_command())

    def _on_offset_spin_value_changed(self, spin_button):
        """Save start position offset."""
        gaupol.conf.preview.offset = spin_button.get_value()

class VideoPage(aeidon.Delegate, gaupol.BuilderDialog):

    """Video preferences page."""

    _widgets = [
        "video_seek_spin",
        "video_subtitle_bg_toggle",
        "video_subtitle_color_button",
        "video_subtitle_font_button",
        "video_time_bg_toggle",
        "video_time_color_button",
        "video_time_font_button",
    ]

    def __init__(self, master, application):
        """Initialize a :class:`PreviewPage` instance."""
        aeidon.Delegate.__init__(self, master)
        self._set_attributes(self._widgets, "video_")
        self.application = application
        self.conf = gaupol.conf.video_player
        self._init_values()

    def _get_custom_font(self, font):
        """Return custom font as string."""
        context = Gtk.Label().get_pango_context()
        font_desc = context.get_font_description()
        custom_font_desc = Pango.FontDescription(font)
        font_desc.merge(custom_font_desc, True)
        return font_desc.to_string()

    def _init_values(self):
        """Initialize default values for widgets."""
        font = self._get_custom_font(self.conf.subtitle_font)
        self._subtitle_font_button.set_font(font)
        font = self._get_custom_font(self.conf.time_font)
        self._time_font_button.set_font(font)
        self._subtitle_bg_toggle.set_active(self.conf.subtitle_background)
        color = gaupol.util.hex_to_rgba(self.conf.subtitle_color)
        color.alpha = self.conf.subtitle_alpha
        self._subtitle_color_button.set_rgba(color)
        self._time_bg_toggle.set_active(self.conf.time_background)
        color = gaupol.util.hex_to_rgba(self.conf.time_color)
        color.alpha = self.conf.time_alpha
        self._time_color_button.set_rgba(color)
        self._seek_spin.set_value(self.conf.seek_length)

    def _on_seek_spin_value_changed(self, spin_button):
        """Save seek length."""
        self.conf.seek_length = spin_button.get_value()

    def _on_subtitle_bg_toggle_toggled(self, toggle_button):
        """Save subtitle background use."""
        self.conf.subtitle_background = toggle_button.get_active()

    def _on_subtitle_color_button_color_set(self, color_button):
        """Save subtitle color."""
        color = color_button.get_rgba()
        self.conf.subtitle_color = gaupol.util.rgba_to_hex(color)
        self.conf.subtitle_alpha = color.alpha

    def _on_subtitle_font_button_font_set(self, font_button):
        """Save subtitle font."""
        self.conf.subtitle_font = font_button.get_font()

    def _on_time_bg_toggle_toggled(self, toggle_button):
        """Save time background use."""
        self.conf.time_background = toggle_button.get_active()

    def _on_time_color_button_color_set(self, color_button):
        """Save time color."""
        color = color_button.get_rgba()
        self.conf.time_color = gaupol.util.rgba_to_hex(color)
        self.conf.time_alpha = color.alpha

    def _on_time_font_button_font_set(self, font_button):
        """Save time font."""
        self.conf.time_font = font_button.get_font()

class PreferencesDialog(gaupol.BuilderDialog):

    """Dialog for editing preferences."""

    def __init__(self, parent, application):
        """Initialize a :class:`PreferencesDialog` instance."""
        # The signal handlers, which live on the page instances created
        # below, are resolved against self already when the UI definition
        # file is parsed. __getattr__ resolves them to placeholders that
        # look up the actual handler from _callbacks only when called.
        self._callbacks = {}
        gaupol.BuilderDialog.__init__(self, "preferences-dialog.ui")
        self._editor_page = EditorPage(self, application)
        self._file_page = FilePage(self, application)
        self._preview_page = PreviewPage(self, application)
        self._video_page = VideoPage(self, application)
        self._callbacks.update(self._get_callbacks())
        self._init_dialog(parent)

    def __getattr__(self, name):
        """Return signal handler placeholder or attribute from dialog."""
        if name.startswith("_on_"):
            # Setting initial widget values in the page constructors emits
            # signals before _callbacks is populated. Those just echo values
            # already in the configuration, so they can be safely dropped.
            def placeholder(*args):
                if name in self._callbacks:
                    return self._callbacks[name](*args)
            return placeholder
        return super().__getattr__(name)

    def _get_callbacks(self):
        """Return a dictionary mapping names to callback methods."""
        callbacks = {}
        for page in (self._editor_page,
                     self._file_page,
                     self._preview_page,
                     self._video_page):

            for name in (x for x in dir(page) if x.startswith("_on_")):
                callbacks[name] = getattr(page, name)
        return callbacks

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.set_default_response(Gtk.ResponseType.CLOSE)
        self.set_transient_for(parent)
        self.set_modal(True)



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Extension delegates of of :class:`gaupol.Application`."""

from .close    import CloseAgent
from .edit     import EditAgent
from .format   import FormatAgent
from .help     import HelpAgent
from .menu     import MenuAgent
from .open     import OpenAgent
from .preview  import PreviewAgent
from .save     import SaveAgent
from .search   import SearchAgent
from .tools    import ToolsAgent
from .update   import UpdateAgent
from .util     import UtilityAgent
from .video    import VideoAgent
from .view     import ViewAgent

classes = (
    CloseAgent,
    EditAgent,
    FormatAgent,
    HelpAgent,
    MenuAgent,
    OpenAgent,
    PreviewAgent,
    SaveAgent,
    SearchAgent,
    ToolsAgent,
    UpdateAgent,
    UtilityAgent,
    VideoAgent,
    ViewAgent,
)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Formatting text."""

import aeidon

class FormatAgent(aeidon.Delegate):

    """Formatting text."""

    @aeidon.deco.export
    def _on_toggle_dialogue_dashes_activate(self, *args):
        """Add or remove dialogue dashes on the selected texts."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.toggle_dialogue_dashes(rows, doc)

    @aeidon.deco.export
    def _on_toggle_italicization_activate(self, *args):
        """Italicize or unitalicize the selected texts."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.toggle_italicization(rows, doc)

    @aeidon.deco.export
    def _on_use_lower_case_activate(self, *args):
        """Change the selected texts to lower case."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.change_case(rows, doc, "lower")

    @aeidon.deco.export
    def _on_use_sentence_case_activate(self, *args):
        """Change the selected texts to sentence case."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.change_case(rows, doc, "capitalize")

    @aeidon.deco.export
    def _on_use_title_case_activate(self, *args):
        """Change the selected texts to title case."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.change_case(rows, doc, "title")

    @aeidon.deco.export
    def _on_use_upper_case_activate(self, *args):
        """Change the selected texts to upper case."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.change_case(rows, doc, "upper")



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Dialog for selecting subtitle files to open."""

import aeidon
import gaupol

from aeidon.i18n   import _
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk
from pathlib import Path

class OpenDialog(Gtk.FileChooserDialog, gaupol.FileDialog):

    """Dialog for selecting subtitle files to open."""

    def __init__(self, parent, title, doc, directory=None):
        """Initialize an :class:`OpenDialog` instance."""
        GObject.GObject.__init__(self)
        self._use_autodetection = aeidon.util.chardet_available()
        self._init_dialog(parent, title)
        self._init_extra_widget()
        self._init_filters()
        self._init_encoding_combo()
        self._init_align_combo()
        self._init_values(doc, directory)

    def _init_align_combo(self):
        """Initialize the align method combo box."""
        store = Gtk.ListStore(str)
        self._align_combo.set_model(store)
        for align_method in aeidon.align_methods:
            store.append((align_method.label,))
        view = self._align_combo.get_child()
        path = gaupol.util.tree_row_to_path(0)
        view.set_displayed_row(path)
        renderer = Gtk.CellRendererText()
        self._align_combo.pack_start(renderer, expand=True)
        self._align_combo.add_attribute(renderer, "text", 0)

    def _init_dialog(self, parent, title):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Open"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_modal(True)
        self.set_title(title)
        self.connect("response", self._on_response)
        self.set_action(Gtk.FileChooserAction.OPEN)

    def _init_extra_widget(self):
        """Initialize the extra widget with encoding and align combos."""
        self._encoding_combo = Gtk.ComboBox(hexpand=True)
        self._encoding_combo.connect(
            "changed", self._on_encoding_combo_changed)
        encoding_label = Gtk.Label(label=_("_Encoding"),
                                   halign=Gtk.Align.START,
                                   use_underline=True)

        encoding_label.set_mnemonic_widget(self._encoding_combo)
        self._align_combo = Gtk.ComboBox(hexpand=True)
        self._align_label = Gtk.Label(label=_("Align _method"),
                                      halign=Gtk.Align.START,
                                      use_underline=True)

        self._align_label.set_mnemonic_widget(self._align_combo)
        grid = Gtk.Grid(row_spacing=6,
                        column_spacing=12,
                        margin_top=12,
                        margin_bottom=12,
                        margin_start=12,
                        margin_end=12)

        grid.attach(encoding_label, 0, 0, 1, 1)
        grid.attach(self._encoding_combo, 1, 0, 1, 1)
        grid.attach(self._align_label, 0, 1, 1, 1)
        grid.attach(self._align_combo, 1, 1, 1, 1)
        # GTK-4 file choosers don't support an extra widget,
        # add to the content area below the file chooser widget.
        self.get_content_area().append(grid)

    def _init_values(self, doc, directory):
        """Initialize default values for widgets."""
        self.set_select_multiple(doc == aeidon.documents.MAIN)
        # Set the location with exactly one call: a second folder load
        # would cancel the first one mid-flight and GTK pops up that
        # cancellation as a modal "Operation was cancelled" error dialog.
        directory = directory or gaupol.conf.file.directory
        if directory and Path(directory).is_dir():
            self.set_current_folder(Gio.File.new_for_path(str(directory)))
        self.set_encoding(gaupol.conf.file.encoding)
        self._align_combo.set_active(gaupol.conf.file.align_method)
        self._align_combo.set_visible(doc == aeidon.documents.TRAN)
        self._align_label.set_visible(doc == aeidon.documents.TRAN)

    def _on_response(self, dialog, response):
        """Save default values for widgets."""
        gaupol.conf.file.encoding = self.get_encoding()
        directory = self.get_current_folder()
        if directory is not None and directory.get_path() is not None:
            gaupol.conf.file.directory = directory.get_path()
        index = self._align_combo.get_active()
        gaupol.conf.file.align_method = aeidon.align_methods[index]



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Miscellaneous functions and decorators."""

import aeidon
import contextlib
import gaupol
import sys
import traceback
import webbrowser

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk

def char_to_px(nchar, font=None):
    """Convert characters to pixels."""
    if nchar < 0: return nchar
    label = Gtk.Label(label="etaoin shrdlu")
    gaupol.style.use_font(label, font)
    width = label.measure(Gtk.Orientation.HORIZONTAL, -1).natural
    return int(round(nchar * width/len(label.props.label)))

def delay_add(delay, function, *args, **kwargs):
    """Call `function` with `args` and `kwargs` once after `delay` (ms)."""
    def call_function(*args, **kwargs):
        function(*args, **kwargs)
        return False # to not be called again.
    return GLib.timeout_add(delay, call_function, *args, **kwargs)

def document_to_text_field(doc):
    """Return :attr:`gaupol.fields` item corresponding to `doc`."""
    if doc == aeidon.documents.MAIN:
        return gaupol.fields.MAIN_TEXT
    if doc == aeidon.documents.TRAN:
        return gaupol.fields.TRAN_TEXT
    raise ValueError(f"Invalid document: {doc!r}")

def flash_dialog(dialog):
    """Run `dialog`, destroy it and return response."""
    response = run_dialog(dialog)
    dialog.destroy()
    return response

def get_content_size(widget, font=None):
    """Return the width and height desired by `widget`."""
    if isinstance(widget, Gtk.TextView):
        return get_text_view_size(widget, font)
    if isinstance(widget, Gtk.TreeView):
        return get_tree_view_size(widget, font)
    raise ValueError(f"Unsupported container type: {type(widget)!r}")

def get_default_player():
    """Return the default video player to use for preview."""
    players = [aeidon.players.MPV, aeidon.players.MPLAYER, aeidon.players.VLC]
    players = [x for x in players if x.found]
    return players[0] if players else aeidon.players.MPV

def get_font():
    """Return custom font or blank string."""
    return (gaupol.conf.editor.custom_font if
            gaupol.conf.editor.use_custom_font and
            gaupol.conf.editor.custom_font else "")

def get_gst_version():
    """Return :mod:`Gst` version number as string or ``None``."""
    try:
        from gi.repository import Gst
        return ".".join(map(str, Gst.version()))
    except Exception:
        return None

def get_libspelling_version():
    """Return libspelling version number as string or ``None``."""
    try:
        from gi.repository import Spelling
        return str(Spelling._version)
    except Exception:
        return None

def get_preview_command():
    """Return command to use for lauching video player."""
    if gaupol.conf.preview.use_custom_command:
        return gaupol.conf.preview.custom_command
    if gaupol.conf.preview.force_utf_8:
        return gaupol.conf.preview.player.command_utf_8
    return gaupol.conf.preview.player.command

def get_text_view_size(text_view, font=None):
    """Return the width and height desired by `text_view`."""
    text_buffer = text_view.get_buffer()
    start, end = text_buffer.get_bounds()
    text = text_buffer.get_text(start, end, False)
    label = Gtk.Label(label=text)
    gaupol.style.use_font(label, font)
    return (label.measure(Gtk.Orientation.HORIZONTAL, -1).natural
            + text_view.get_left_margin()
            + text_view.get_right_margin(),
            label.measure(Gtk.Orientation.VERTICAL, -1).natural)

def get_tree_view_size(tree_view, font=None):
    """Return the width and height desired by `tree_view`."""
    scroller = tree_view.get_parent()
    policy = scroller.get_policy()
    scroller.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
    width = scroller.measure(Gtk.Orientation.HORIZONTAL, -1).natural
    height = scroller.measure(Gtk.Orientation.VERTICAL, -1).natural
    scroller.set_policy(*policy)
    return width, height

@aeidon.deco.once
def get_zebra_color(tree_view):
    """Return background color to use for tree view zebra-stripes."""
    # XXX: Zebra stripes would be faster and cleaner done with CSS
    # selectors :nth-child(odd) and :nth-child(even), but they don't
    # work: GtkTreeView has no per-row CSS nodes, the selectors match
    # the tree view widget itself among its siblings (tested 7/2026).
    # https://bugzilla.gnome.org/show_bug.cgi?id=709617#c1
    # Foreground color at low alpha composited over the background
    # gives a subtle stripe on any theme, light or dark.
    color = tree_view.get_color()
    color.alpha = 0.05
    return color

@aeidon.deco.once
def gst_available():
    """Return ``True`` if :mod:`Gst` and needed plugins are available."""
    try:
        from gi.repository import Gst
    except Exception:
        return False
    if not Gst.ElementFactory.find("playbin"):
        print("GStreamer found, but playbin missing.",
              "Try installing gst-plugins-base.",
              file=sys.stderr)
        return False
    if not Gst.ElementFactory.find("textoverlay"):
        print("GStreamer found, but textoverlay missing.",
              "Try installing gst-plugins-base.",
              file=sys.stderr)
        return False
    if not Gst.ElementFactory.find("timeoverlay"):
        print("GStreamer found, but timeoverlay missing.",
              "Try installing gst-plugins-base.",
              file=sys.stderr)
        return False
    if not Gst.ElementFactory.find("gtk4paintablesink"):
        print("GStreamer found, but gtk4paintablesink missing.",
              "Try installing gst-plugins-rs.",
              file=sys.stderr)
        return False
    if Gst.ElementFactory.find("vaapisink"):
        print("GStreamer-vaapi found, known to sometimes cause trouble.",
              "If video doesn't play, please try uninstalling it.",
              "See https://github.com/otsaloma/gaupol/issues/79",
              file=sys.stderr)
    return True

def hex_to_rgba(string):
    """Return a :class:`Gdk.RGBA` for hexadecimal `string`."""
    rgba = Gdk.RGBA()
    success = rgba.parse(string)
    if not success:
        raise ValueError(f"Parsing string {string!r} failed")
    return rgba

def idle_add(function, *args, **kwargs):
    """Call `function` with `args` and `kwargs` when idle."""
    def call_function(*args, **kwargs):
        function(*args, **kwargs)
        return False # to not be called again.
    return GLib.idle_add(call_function, *args, **kwargs)

def iterate_main():
    """Iterate the GLib main loop while events are pending."""
    context = GLib.MainContext.default()
    while context.pending():
        context.iteration(False)

def lines_to_px(nlines, font=None):
    """Convert lines to pixels."""
    if nlines < 0: return nlines
    text = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    label = Gtk.Label(label=text)
    gaupol.style.use_font(label, font)
    height = label.measure(Gtk.Orientation.VERTICAL, -1).natural
    return int(round(nlines * height))

def new_hbox(spacing):
    """Return a new horizontal :class:`Gtk.Box`."""
    return Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL,
                   spacing=spacing)

def new_vbox(spacing):
    """Return a new vertical :class:`Gtk.Box`."""
    return Gtk.Box(orientation=Gtk.Orientation.VERTICAL,
                   spacing=spacing)

def pack_start(box, widget):
    """Pack widget to box without expand."""
    box.append(widget)

def pack_start_expand(box, widget):
    """Pack widget to box with expand."""
    if box.get_orientation() == Gtk.Orientation.HORIZONTAL:
        widget.set_hexpand(True)
    else:
        widget.set_vexpand(True)
    box.append(widget)

def prepare_text_view(text_view):
    """Set spell-check, line-length margin and font properties."""
    if (gaupol.SpellChecker.available() and
        gaupol.conf.spell_check.inline):
        language = gaupol.conf.spell_check.language
        with contextlib.suppress(Exception):
            checker = gaupol.SpellChecker(language)
            checker.attach(text_view)
    connect = gaupol.conf.editor.connect
    def update_margin(section, value, text_view):
        if gaupol.conf.editor.show_lengths_edit:
            return gaupol.ruler.connect_text_view(text_view)
        return gaupol.ruler.disconnect_text_view(text_view)
    connect("notify::show_lengths_edit", update_margin, text_view)
    update_margin(None, None, text_view)
    gaupol.style.use_font(text_view, "custom")

def raise_default(expression):
    """Raise :exc:`gaupol.Default` if `expression` evaluates to ``True``."""
    if expression:
        raise gaupol.Default

def rgba_to_hex(color):
    """Return hexadecimal string for :class:`Gdk.RGBA` `color`."""
    red = int(color.red * 255)
    green = int(color.green * 255)
    blue = int(color.blue * 255)
    return f"#{red:02x}{green:02x}{blue:02x}"

def run_dialog(dialog):
    """
    Run `dialog` in a nested main loop and return response.

    A stand-in for ``Gtk.Dialog.run``, which was removed in GTK-4. Blocks in a
    nested main loop until the "response" signal, keeping callers simple,
    synchronous code. Closing the dialog window emits response
    ``Gtk.ResponseType.DELETE_EVENT``.
    """
    response = Gtk.ResponseType.NONE
    loop = GLib.MainLoop()
    def on_response(dialog, response_id):
        nonlocal response
        response = response_id
        loop.quit()
    def on_unrealize(dialog):
        # Don't hang if dialog is destroyed without a response.
        loop.quit()
    response_id = dialog.connect("response", on_response)
    unrealize_id = dialog.connect("unrealize", on_unrealize)
    dialog.present()
    loop.run()
    for handler in [response_id, unrealize_id]:
        if dialog.handler_is_connected(handler):
            dialog.disconnect(handler)
    return response

def scale_to_content(widget, min_nchar=0,  max_nchar=32768,
                     min_nlines=0, max_nlines=32768, font=None):
    """Set `widget's` size by content, but limited by `min` and `max`."""
    width, height = get_content_size(widget, font)
    width  = max(width, char_to_px(min_nchar, font))
    width  = min(width, char_to_px(max_nchar, font))
    height = max(height, lines_to_px(min_nlines, font))
    height = min(height, lines_to_px(max_nlines, font))
    parent = widget.get_parent()
    if isinstance(parent, Gtk.ScrolledWindow):
        # Vaguely account for possible scrollbars.
        return parent.set_size_request(width + 24, height + 24)
    widget.set_size_request(width, height)

def scale_to_size(widget, nchar, nlines, font=None):
    """Set `widget`'s size to `nchar` and `nlines`."""
    width  = char_to_px(nchar, font)
    height = lines_to_px(nlines, font)
    parent = widget.get_parent()
    if isinstance(parent, Gtk.ScrolledWindow):
        # Vaguely account for possible scrollbars.
        return parent.set_size_request(width + 24, height + 24)
    widget.set_size_request(width, height)

def separate_combo(store, itr, data=None):
    """Separator function for combo box models."""
    return store.get_value(itr, 0) == gaupol.COMBO_SEPARATOR

def set_cursor_busy(window):
    """Set mouse pointer busy when above window."""
    window.set_cursor_from_name("wait")
    iterate_main()

def set_cursor_normal(window):
    """Set mouse pointer normal when above window."""
    window.set_cursor_from_name("default")
    iterate_main()

def set_widget_margins(widget, margin):
    """Set equal margins on all sides of `widget`."""
    widget.set_margin_top(margin)
    widget.set_margin_bottom(margin)
    widget.set_margin_start(margin)
    widget.set_margin_end(margin)

def show_exception(exctype, exc, tb):
    """A :class:`gaupol.DebugDialog` :attr`sys.excepthook`."""
    traceback.print_exception(exctype, exc, tb)
    if not isinstance(exc, Exception): return
    try: # to avoid recursion.
        manager = getattr(gaupol, "appman", None)
        parent = manager.get_active_window() if manager else None
        dialog = gaupol.DebugDialog(parent)
        dialog.set_text(exctype, exc, tb)
        response = flash_dialog(dialog)
        if response == Gtk.ResponseType.NO:
            raise SystemExit(1)
    except Exception:
        traceback.print_exc()

def show_uri(uri):
    """Open `uri` in default application."""
    try:
        return Gio.AppInfo.launch_default_for_uri(uri)
    except Exception:
        # Launching fails on Windows and some misconfigured installations.
        # GError: No application is registered as handling this file
        # Operation not supported
        if uri.startswith(("http://", "https://")):
            return webbrowser.open(uri)
        raise # Exception

def text_field_to_document(field):
    """Return :attr:`aeidon.documents` item corresponding to `field`."""
    if field == gaupol.fields.MAIN_TEXT:
        return aeidon.documents.MAIN
    if field == gaupol.fields.TRAN_TEXT:
        return aeidon.documents.TRAN
    raise ValueError(f"Invalid field: {field!r}")

def tree_path_to_row(path):
    """
    Convert `path` to a list row integer.

    `path` can be either a :class:`Gtk.Treepath` instance or a string
    representation of it (as commonly used by various callbacks).
    """
    if path is None: return None
    if isinstance(path, Gtk.TreePath):
        return path.get_indices()[0]
    if isinstance(path, str):
        return int(path)
    raise TypeError(f"Bad type {type(path)!r} for path {path!r}")

def tree_row_to_path(row):
    """Convert list row integer to a :class:`Gtk.TreePath`."""
    if row is None: return None
    return Gtk.TreePath.new_from_string(str(row))



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""User-activatable actions for :class:`gaupol.Application`."""

from gaupol.actions import audio
from gaupol.actions import edit
from gaupol.actions import file
from gaupol.actions import help
from gaupol.actions import projects
from gaupol.actions import text
from gaupol.actions import tools
from gaupol.actions import video
from gaupol.actions import view

classes = (
    audio.SetAudioLanguageAction,
    audio.VolumeDownAction,
    audio.VolumeUpAction,
    edit.EditPreferencesAction,
    edit.EditNextValueAction,
    edit.EditValueAction,
    edit.EndEarlierAction,
    edit.EndLaterAction,
    edit.ExtendSelectionToBeginningAction,
    edit.ExtendSelectionToEndAction,
    edit.InsertSubtitleAtVideoPositionAction,
    edit.InsertSubtitlesAction,
    edit.InvertSelectionAction,
    edit.MergeSubtitlesAction,
    edit.RedoActionAction,
    edit.RemoveSubtitlesAction,
    edit.SelectAllAction,
    edit.SelectNextFromVideoPositionAction,
    edit.SelectPreviousFromVideoPositionAction,
    edit.SetEndFromVideoPositionAction,
    edit.SetStartFromVideoPositionAction,
    edit.SplitSubtitleAction,
    edit.StartEarlierAction,
    edit.StartLaterAction,
    edit.UndoActionAction,
    file.CloseProjectAction,
    file.NewProjectAction,
    file.OpenMainFilesAction,
    file.OpenTranslationFileAction,
    file.QuitAction,
    file.SaveMainDocumentAction,
    file.SaveMainDocumentAsAction,
    file.SaveTranslationDocumentAction,
    file.SaveTranslationDocumentAsAction,
    help.BrowseDocumentationAction,
    help.ReportABugAction,
    help.ViewAboutDialogAction,
    projects.ActivateNextProjectAction,
    projects.ActivatePreviousProjectAction,
    projects.ActivateProjectAction,
    projects.CloseAllProjectsAction,
    projects.SaveAllDocumentsAction,
    projects.SaveAllDocumentsAsAction,
    text.ClearTextsAction,
    text.CopyTextsAction,
    text.CutTextsAction,
    text.FindAndReplaceAction,
    text.FindNextAction,
    text.FindPreviousAction,
    text.PasteTextsAction,
    text.ToggleDialogDashesAction,
    text.ToggleItalicizationAction,
    text.UseLowerCaseAction,
    text.UseSentenceCaseAction,
    text.UseTitleCaseAction,
    text.UseUpperCaseAction,
    tools.AdjustDurationsAction,
    tools.AppendFileAction,
    tools.CheckSpellingAction,
    tools.ConfigureSpellCheckAction,
    tools.ConvertFramerateAction,
    tools.CorrectTextsAction,
    tools.PreviewAction,
    tools.SelectVideoFileAction,
    tools.ShiftPositionsAction,
    tools.SplitProjectAction,
    tools.TransformPositionsAction,
    video.LoadVideoAction,
    video.PlayPauseAction,
    video.PlaySelectionAction,
    video.SeekBackwardAction,
    video.SeekForwardAction,
    video.SeekNextAction,
    video.SeekPreviousAction,
    video.SeekSelectionEndAction,
    video.SeekSelectionStartAction,
    view.SetEditModeAction,
    view.SetFramerateAction,
    view.SetLayoutAction,
    view.ToggleDurationColumnAction,
    view.ToggleEndColumnAction,
    view.ToggleMainTextColumnAction,
    view.ToggleNumberColumnAction,
    view.TogglePlayerAction,
    view.ToggleStartColumnAction,
    view.ToggleTranslationTextColumnAction,
)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Cell renderer for time data in format ``[-]HH:MM:SS.SSS``."""

import gaupol

from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk

class TimeCellRenderer(Gtk.CellRendererText):

    """Cell renderer for time data in format ``[-]HH:MM:SS.SSS``."""

    __gtype_name__ = "TimeCellRenderer"

    def __init__(self):
        """Initialize a :class:`TimeCellRenderer` instance."""
        GObject.GObject.__init__(self)

    def do_start_editing(self, event, widget, path, bg_area, cell_area, flags):
        """Initialize and return a :class:`gaupol.TimeEntry` widget."""
        editor = gaupol.TimeEntry()
        gaupol.style.use_font(editor, "custom")
        # Match the renderer's yalign=0 when the cell
        # is taller than one line of text.
        editor.set_valign(Gtk.Align.START)
        editor.set_has_frame(False)
        editor.set_alignment(self.props.xalign)
        editor.set_text(self.props.text)
        editor.select_region(0, -1)
        editor.gaupol_path = path
        controller = Gtk.EventControllerFocus()
        controller.connect("leave", self._on_editor_focus_leave, editor)
        editor.add_controller(controller)
        controller = Gtk.EventControllerKey()
        controller.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        controller.connect("key-pressed", self._on_editor_key_pressed, editor)
        editor.add_controller(controller)
        return editor

    def _on_editor_focus_leave(self, controller, editor):
        """End editing."""
        # The leave event is delivered in the middle of the focus
        # change and removing the focused editor there sends GTK's
        # focus handling into an infinite loop. Defer to idle and
        # skip if editing already ended (editor unparented).
        def end_editing():
            if editor.get_parent() is None: return
            editor.remove_widget()
            self.emit("editing-canceled")
        gaupol.util.idle_add(end_editing)

    def _on_editor_key_pressed(self, controller, keyval, keycode, state, editor):
        """End editing if ``Enter`` or ``Escape`` pressed."""
        if (state &
            (Gdk.ModifierType.SHIFT_MASK |
             Gdk.ModifierType.CONTROL_MASK)): return False
        if keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter):
            editor.remove_widget()
            self.emit("edited", editor.gaupol_path, editor.get_text())
            return True
        if keyval == Gdk.KEY_Escape:
            editor.remove_widget()
            self.emit("editing-canceled")
            return True
        return False



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Dialog for selecting a subtitle file to save."""

import aeidon
import gaupol

from aeidon.i18n   import _
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk
from pathlib import Path

class SaveDialog(Gtk.FileChooserDialog, gaupol.FileDialog):

    """Dialog for selecting a subtitle file to save."""

    def __init__(self, parent, title, mode, path=None):
        """Initialize a :class:`SaveDialog` instance."""
        GObject.GObject.__init__(self)
        self._mode = mode
        self._init_dialog(parent, title)
        self._init_extra_widget()
        self._init_filters()
        self._init_format_combo()
        self._init_encoding_combo()
        self._init_newline_combo()
        self._init_framerate_combo()
        self._init_values(path)

    def get_format(self):
        """Return the selected format."""
        index = self._format_combo.get_active()
        return aeidon.formats[index]

    def get_framerate(self):
        """Return the selected framerate."""
        index = self._framerate_combo.get_active()
        return aeidon.framerates[index]

    def get_newline(self):
        """Return the selected newline."""
        index = self._newline_combo.get_active()
        return aeidon.newlines[index]

    def _init_dialog(self, parent, title):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Save"), Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_modal(True)
        self.set_title(title)
        self.connect("response", self._on_response)
        # Add a possibly missing filename extension before GTK's
        # overwrite confirmation runs on save button click.
        save_button = self.get_widget_for_response(Gtk.ResponseType.OK)
        gesture = Gtk.GestureClick()
        gesture.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        gesture.connect("pressed", self._on_save_button_pressed)
        save_button.add_controller(gesture)
        self.set_action(Gtk.FileChooserAction.SAVE)

    def _init_extra_widget(self):
        """Initialize the extra widget with format etc. combos."""
        self._format_combo = Gtk.ComboBox(hexpand=True)
        self._format_combo.connect("changed", self._on_format_combo_changed)
        format_label = Gtk.Label(label=_("For_mat"),
                                 halign=Gtk.Align.START,
                                 use_underline=True)

        format_label.set_mnemonic_widget(self._format_combo)
        self._encoding_combo = Gtk.ComboBox(hexpand=True)
        self._encoding_combo.connect(
            "changed", self._on_encoding_combo_changed)
        encoding_label = Gtk.Label(label=_("_Encoding"),
                                   halign=Gtk.Align.START,
                                   use_underline=True)

        encoding_label.set_mnemonic_widget(self._encoding_combo)
        self._newline_combo = Gtk.ComboBox(hexpand=True)
        newline_label = Gtk.Label(label=_("Ne_wlines"),
                                  halign=Gtk.Align.START,
                                  use_underline=True)

        newline_label.set_mnemonic_widget(self._newline_combo)
        self._framerate_combo = Gtk.ComboBox(hexpand=True)
        self._framerate_label = Gtk.Label(label=_("F_ramerate"),
                                          halign=Gtk.Align.START,
                                          use_underline=True)

        self._framerate_label.set_mnemonic_widget(self._framerate_combo)
        grid = Gtk.Grid(row_spacing=6,
                        column_spacing=12,
                        margin_top=12,
                        margin_bottom=12,
                        margin_start=12,
                        margin_end=12)

        grid.attach(format_label, 0, 0, 1, 1)
        grid.attach(self._format_combo, 1, 0, 1, 1)
        grid.attach(encoding_label, 0, 1, 1, 1)
        grid.attach(self._encoding_combo, 1, 1, 1, 1)
        grid.attach(newline_label, 0, 2, 1, 1)
        grid.attach(self._newline_combo, 1, 2, 1, 1)
        grid.attach(self._framerate_label, 0, 3, 1, 1)
        grid.attach(self._framerate_combo, 1, 3, 1, 1)
        # GTK-4 file choosers don't support an extra widget,
        # add to the content area below the file chooser widget.
        self.get_content_area().append(grid)

    def _init_format_combo(self):
        """Initialize the format combo box."""
        store = Gtk.ListStore(str)
        self._format_combo.set_model(store)
        for name in (x.label for x in aeidon.formats):
            if name == "SubRip":
                # Mark the SubRip format as recommended, since
                # it's the most common one and best supported,
                # both in Gaupol and elsewhere.
                name = _("{} (recommended)").format(name)
            store.append((name,))
        view = self._format_combo.get_child()
        path = gaupol.util.tree_row_to_path(0)
        view.set_displayed_row(path)
        renderer = Gtk.CellRendererText()
        self._format_combo.pack_start(renderer, expand=True)
        self._format_combo.add_attribute(renderer, "text", 0)

    def _init_framerate_combo(self):
        """Initialize the framerate combo box."""
        store = Gtk.ListStore(str)
        self._framerate_combo.set_model(store)
        for name in (x.label for x in aeidon.framerates):
            store.append((name,))
        view = self._framerate_combo.get_child()
        path = gaupol.util.tree_row_to_path(0)
        view.set_displayed_row(path)
        renderer = Gtk.CellRendererText()
        self._framerate_combo.pack_start(renderer, expand=True)
        self._framerate_combo.add_attribute(renderer, "text", 0)

    def _init_newline_combo(self):
        """Initialize the newline combo box."""
        store = Gtk.ListStore(str)
        self._newline_combo.set_model(store)
        for name in (x.label for x in aeidon.newlines):
            store.append((name,))
        view = self._newline_combo.get_child()
        path = gaupol.util.tree_row_to_path(0)
        view.set_displayed_row(path)
        renderer = Gtk.CellRendererText()
        self._newline_combo.pack_start(renderer, expand=True)
        self._newline_combo.add_attribute(renderer, "text", 0)

    def _init_values(self, path):
        """Initialize default values for widgets."""
        # Set the location with exactly one load: a second one (set_file
        # or set_current_folder) would cancel the first one mid-flight
        # and GTK pops up that cancellation as a modal "Operation was
        # cancelled" error dialog. set_current_name is not a load.
        if path is not None and Path(path).is_file():
            self.set_file(Gio.File.new_for_path(str(path)))
        else:
            if path is not None:
                self.set_current_name(Path(path).name)
            if (gaupol.conf.file.directory and
                Path(gaupol.conf.file.directory).is_dir()):
                directory = Gio.File.new_for_path(gaupol.conf.file.directory)
                self.set_current_folder(directory)
        self.set_encoding(gaupol.conf.file.encoding)
        self.set_format(gaupol.conf.file.format)
        self.set_newline(gaupol.conf.file.newline)
        self.set_framerate(gaupol.conf.editor.framerate)
        self._framerate_combo.set_visible(False)
        self._framerate_label.set_visible(False)

    def _on_format_combo_changed(self, *args):
        """Change the extension of the current filename."""
        format = self.get_format()
        basename = self.get_current_name()
        if basename and not basename.endswith(format.extension):
            basename = aeidon.util.replace_extension(basename, format)
            self.set_current_name(str(basename))
        visible = (format.mode != self._mode)
        self._framerate_combo.set_visible(visible)
        self._framerate_label.set_visible(visible)

    def _on_response(self, dialog, response):
        """Save default values for widgets."""
        directory = self.get_current_folder()
        if directory is not None and directory.get_path() is not None:
            gaupol.conf.file.directory = directory.get_path()
        gaupol.conf.file.encoding = self.get_encoding()
        gaupol.conf.file.format = self.get_format()
        gaupol.conf.file.newline = self.get_newline()
        gaupol.conf.editor.framerate = self.get_framerate()
        if response == Gtk.ResponseType.OK:
            # If the filename is lacking the extension, add it. This already
            # happened in self._on_save_button_pressed, before overwrite
            # confirmation, if the save button was clicked, but not
            # necessarily if the user hit Enter on the keyboard.
            self._format_combo.emit("changed")

    def _on_save_button_pressed(self, gesture, n_press, x, y):
        """Ensure that the filename contains an extension."""
        # Add possibly lacking extension to the filename.
        self._format_combo.emit("changed")

    def set_format(self, format):
        """Set the selected format."""
        if format is None: return
        self._format_combo.set_active(format)

    def set_framerate(self, framerate):
        """Set the selected framerate."""
        if framerate is None: return
        self._framerate_combo.set_active(framerate)

    def set_newline(self, newline):
        """Set the selected newline."""
        if newline is None: return
        self._newline_combo.set_active(newline)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""GTK user interface controller for :class:`aeidon.Project`."""

import aeidon
import contextlib
import gaupol
import itertools
import sys

from aeidon.i18n   import _
from gi.repository import Gdk
from gi.repository import Gtk

class ApplicationMeta(type):

    """
    Application metaclass with delegated methods added.

    Public methods are added to the class dictionary during :meth:`__new__`
    in order to fool Sphinx (and perhaps other API documentation generators)
    into thinking that the resulting instantiated class actually contains those
    methods, which it does not since the methods are removed during
    :meth:`Application.__init__`.
    """

    def __new__(meta, class_name, bases, dic):
        new_dict = dic.copy()
        for agent_class in gaupol.agents.classes:
            def is_delegate_method(name):
                value = getattr(agent_class, name)
                return (callable(value) and
                        hasattr(value, "export") and
                        value.export is True)

            attr_names = list(filter(is_delegate_method, dir(agent_class)))
            for attr_name in attr_names:
                new_dict[attr_name] = getattr(agent_class, attr_name)
        return type.__new__(meta, class_name, bases, new_dict)

class Application(aeidon.Observable, metaclass=ApplicationMeta):

    """
    GTK user interface controller for :class:`aeidon.Project`.

    :ivar clipboard: Instance of :class:`aeidon.Clipboard` used
    :ivar counter: Iterator used for naming unsaved documents
    :ivar _delegations: Dictionary mapping method names to agent methods
    :ivar notebook: A :class:`Gtk.Notebook` used to hold multiple projects
    :ivar pages: List of :class:`gaupol.Page` currently open
    :ivar paned: A :class:`Gtk.Paned` to hold player and subtitles
    :ivar pattern: Last used search pattern or blank if not used
    :ivar play_button: The play/pause button on the video player toolbar
    :ivar player: A :class:`gaupol.VideoPlayer` instance or ``None``
    :ivar player_box: Box containing video player etc.
    :ivar player_toolbar: A toolbar :class:`Gtk.Box` for video player actions
    :ivar recent_manager: Instance of :class:`Gtk.RecentManager` used
    :ivar replacement: Last used search replacement or blank if not used
    :ivar seekbar: Video player seekbar (a :class:`Gtk.Scale` instance)
    :ivar toast: Instance of :class:`gaupol.Toast` used
    :ivar volume_button: A :class:`Gtk.VolumeButton` in the player toolbar
    :ivar window: A :class:`Gtk.ApplicationWindow` used to hold all the widgets
    :ivar x_clipboard: A :class:`Gdk.Clipboard` used for desktop-wide copying

    Signals and their arguments for callback functions:
     * ``init-done``: application
     * ``page-added``: application, page
     * ``page-changed``: application, page
     * ``page-closed``: application, page
     * ``page-saved``: application, page
     * ``page-switched``: application, page
     * ``pages-reordered``: application, page, number
     * ``quit``: application
    """

    signals = (
        "init-done",
        "page-added",
        "page-changed",
        "page-closed",
        "page-saved",
        "page-switched",
        "pages-reordered",
        "quit",
    )

    def __init__(self):
        """Initialize an :class:`Application` instance."""
        aeidon.Observable.__init__(self)
        self.clipboard = aeidon.Clipboard()
        self.counter = itertools.count(1)
        self._delegations = {}
        self.notebook = None
        self.pages = []
        self.pattern = ""
        self.play_button = None
        self.player = None
        self.player_box = None
        self.player_toolbar = None
        self.recent_manager = Gtk.RecentManager.get_default()
        self.replacement = ""
        self.seekbar = None
        self.toast = None
        self.volume_button = None
        self.window = None
        self.x_clipboard = None
        self._init_delegations()
        self._init_custom_framerates()
        self._init_gui()
        self.update_gui()
        self.window.present()
        self.emit("init-done")

    def __getattr__(self, name):
        """Return method delegated to an agent."""
        try:
            return self._delegations[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        """Set value of attribute `name`."""
        return aeidon.Observable.__setattr__(self, name, value)

    def _init_actions(self):
        """Initialize user-activatable actions."""
        for action_class in gaupol.actions.classes:
            action = action_class()
            name = f"win.{action.props.name}"
            if hasattr(gaupol, "appman"):
                gaupol.appman.set_accels_for_action(
                    name, action.accelerators)
            # Route all activations through a common handler that
            # re-checks doability, keeping the real callback aside.
            stub = action.props.name.replace("-", "_")
            action.callback = getattr(self, f"_on_{stub}_activate")
            action.connect("activate", self._on_action_activate)
            self.window.add_action(action)

    def _init_custom_framerates(self):
        """Add framerates from ``conf.editor.custom_framerates``."""
        menu = self.get_menubar_section("custom-framerates-placeholder")
        for value in sorted(gaupol.conf.editor.custom_framerates):
            name = f"FPS_{value:.3f}".replace(".", "_")
            if hasattr(aeidon.framerates, name): continue
            setattr(aeidon.framerates, name, aeidon.EnumerationItem())
            framerate = getattr(aeidon.framerates, name)
            framerate.label = _("{:.3f} fps").format(value)
            framerate.value = float(value)
            if menu is not None:
                # Menubar not available when running unit tests.
                action = f"win.set-framerate::{name}"
                menu.append(framerate.label, action)

    def _init_delegations(self):
        """Initialize the delegation mappings."""
        for agent_class in gaupol.agents.classes:
            agent = agent_class(self)
            def is_delegate_method(name):
                value = getattr(agent, name)
                return (callable(value) and
                        hasattr(value, "export") and
                        value.export is True)

            attr_names = list(filter(is_delegate_method, dir(agent)))
            for attr_name in attr_names:
                attr_value = getattr(agent, attr_name)
                if attr_name in self._delegations:
                    raise ValueError(f"Multiple definitions of {attr_name!r}")

                self._delegations[attr_name] = attr_value
                # Remove class-level function added by ApplicationMeta.
                if hasattr(self.__class__, attr_name):
                    delattr(self.__class__, attr_name)

    def _init_gui(self):
        """Initialize the user interface."""
        vbox = gaupol.util.new_vbox(spacing=0)
        self._init_x_clipboard()
        self._init_window()
        self._init_actions()
        self._init_header_bar()
        self._init_paned(vbox)
        self._init_player_box(self.paned)
        self._init_notebook(self.paned)
        self.window.set_child(vbox)
        self._init_visibilities()

    def _init_header_bar(self):
        """Initialize the window header bar."""
        header = Gtk.HeaderBar()
        # Open split button: win.open-main-files
        # plus a menu listing recent files.
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        box.add_css_class("linked")
        button = Gtk.Button(icon_name="document-open-symbolic")
        button.set_action_name("win.open-main-files")
        button.set_tooltip_text(_("Open main files"))
        box.append(button)
        button = Gtk.MenuButton(icon_name="pan-down-symbolic")
        button.set_menu_model(self.get_menubar_section("open-recent-main-placeholder"))
        button.set_tooltip_text(_("Open a recently used main file"))
        box.append(button)
        header.pack_start(box)
        # win.save-main-document
        button = Gtk.Button(icon_name="document-save-symbolic")
        button.set_action_name("win.save-main-document")
        button.set_tooltip_text(_("Save the current main document"))
        header.pack_start(button)
        # win.undo-action and win.redo-action as a group
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        box.add_css_class("linked")
        button = Gtk.Button(icon_name="edit-undo-symbolic")
        button.set_action_name("win.undo-action")
        button.set_tooltip_text(_("Undo the last action"))
        box.append(button)
        button = Gtk.Button(icon_name="edit-redo-symbolic")
        button.set_action_name("win.redo-action")
        button.set_tooltip_text(_("Redo the last undone action"))
        box.append(button)
        header.pack_start(box)
        # win.preview
        button = Gtk.Button(icon_name="media-playback-start-symbolic")
        button.set_action_name("win.preview")
        button.set_tooltip_text(_("Preview from selected position with a video player"))
        header.pack_end(button)
        # win.find-and-replace
        button = Gtk.Button(icon_name="edit-find-symbolic")
        button.set_action_name("win.find-and-replace")
        button.set_tooltip_text(_("Search for and replace text"))
        header.pack_end(button)
        self.window.set_titlebar(header)

    def _init_notebook(self, paned):
        """Initialize the notebook."""
        self.notebook = Gtk.Notebook()
        self.notebook.set_scrollable(True)
        self.notebook.set_show_border(False)
        self.notebook.set_show_tabs(False)
        target = Gtk.DropTarget.new(Gdk.FileList, Gdk.DragAction.COPY)
        target.connect("drop", self._on_notebook_drop)
        self.notebook.add_controller(target)
        aeidon.util.connect(self, "notebook", "page-reordered")
        callback = self._on_notebook_switch_page
        self.notebook.connect_after("switch-page", callback)
        overlay = Gtk.Overlay()
        overlay.set_child(self.notebook)
        self.toast = gaupol.Toast()
        overlay.add_overlay(self.toast)
        paned.set_end_child(overlay)

    def _init_paned(self, vbox):
        """Intialize the paned layout."""
        orientation = gaupol.conf.application_window.layout
        self.paned = Gtk.Paned(orientation=orientation)
        gaupol.util.pack_start_expand(vbox, self.paned)

    def _init_player_box(self, paned):
        """Initialize the video player horizontal box."""
        # This will actually be added to paned once a video is loaded.
        layout = gaupol.conf.application_window.layout
        if layout == Gtk.Orientation.HORIZONTAL:
            self.player_box = gaupol.util.new_vbox(spacing=0)
        if layout == Gtk.Orientation.VERTICAL:
            self.player_box = gaupol.util.new_hbox(spacing=0)
        # Keep hidden until a video is loaded. Tab visibility logic
        # in UpdateAgent._update_widgets depends on this too.
        self.player_box.set_visible(False)

    def _init_visibilities(self):
        """Initialize visibilities of hideable widgets."""
        self.show_message(None)

    def _init_window(self):
        """Initialize the main window."""
        self.window = Gtk.ApplicationWindow(
            application=getattr(gaupol, "appman", None))
        self.window.set_show_menubar(True)
        self.window.set_icon_name("io.otsaloma.gaupol")
        Gtk.Window.set_default_icon_name("io.otsaloma.gaupol")
        self.window.set_default_size(*gaupol.conf.application_window.size)
        if gaupol.conf.application_window.maximized:
            self.window.maximize()
        aeidon.util.connect(self, "window", "close-request")
        aeidon.util.connect(self, "window", "notify::maximized")
        gaupol.style.load_css(self.window)

    def _init_x_clipboard(self):
        """Initialize the desktop-wide clipboard."""
        self.x_clipboard = Gdk.Display.get_default().get_clipboard()

    def _on_action_activate(self, action, *args):
        """Run `action`'s callback if it is currently doable."""
        # Action sensitivity can in rare corner cases go stale.
        # Re-check the action is doable before running the callback.
        page = self.get_current_page()
        rows = []
        with contextlib.suppress(AttributeError):
            rows = page.view.get_selected_rows()
        if action.update_enabled(self, page, rows):
            return action.callback(action, *args)
        else:
            print(f"{action.props.name} is no longer doable", file=sys.stderr)
            return None



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Dialog for configuring spell-check."""

import aeidon
import contextlib
import gaupol

from aeidon.i18n   import _
from gi.repository import Gtk

class LanguageDialog(gaupol.BuilderDialog):

    """Dialog for configuring spell-check."""

    _widgets = [
        "all_radio",
        "content_box",
        "current_radio",
        "language_list",
        "language_title_label",
        "main_radio",
        "scroller",
        "target_vbox",
        "tran_radio",
    ]

    def __init__(self, parent, show_target=True):
        """Initialize a :class:`LanguageDialog` instance."""
        gaupol.BuilderDialog.__init__(self, "language-dialog.ui")
        self._init_dialog(parent)
        self._init_visibilities(show_target)
        self._init_language_list()
        self._init_values()
        self._init_scroller(show_target)

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.set_default_response(Gtk.ResponseType.CLOSE)
        self.set_transient_for(parent)
        self.set_modal(True)

    def _init_language_list(self):
        """Initialize the list of languages."""
        self._language_list.connect("row-activated", self._on_language_list_row_activated)
        locales = []
        with contextlib.suppress(Exception):
            locales = aeidon.SpellChecker.list_languages()
        group = None
        for locale in locales:
            name = locale
            with contextlib.suppress(Exception):
                name = aeidon.locales.code_to_name(locale)
            # Arrow keys in a list box move focus between rows, not between
            # radio buttons, so keep the radio buttons out of the focus chain
            # and let the rows handle keyboard focus and activation.
            radio = Gtk.CheckButton(label=name, hexpand=True, can_focus=False, group=group)
            radio.set_active(locale == gaupol.conf.spell_check.language)
            radio.connect("toggled", self._on_language_radio_toggled)
            radio.gaupol_locale = locale
            group = group or radio
            row = Gtk.ListBoxRow(activatable=True, selectable=False)
            row.set_child(radio)
            row.gaupol_radio = radio
            self._language_list.append(row)

    def _init_scroller(self, show_target):
        """Set the scroller height from the target column."""
        if show_target:
            height  = self._target_vbox.measure(Gtk.Orientation.VERTICAL, -1).natural
            height += self._content_box.get_margin_top()
            height += self._content_box.get_margin_bottom()
            self._scroller.set_max_content_height(height)
        else:
            height = gaupol.util.lines_to_px(25)
            self._scroller.set_max_content_height(height)

    def _init_values(self):
        """Initialize default values for widgets."""
        field = gaupol.conf.spell_check.field
        target = gaupol.conf.spell_check.target
        self._main_radio.set_active(field == gaupol.fields.MAIN_TEXT)
        self._tran_radio.set_active(field == gaupol.fields.TRAN_TEXT)
        self._all_radio.set_active(target == gaupol.targets.ALL)
        self._current_radio.set_active(target == gaupol.targets.CURRENT)

    def _init_visibilities(self, show_target):
        """Initialize visibilities of target widgets."""
        if not show_target:
            self._language_title_label.set_visible(False)
            self._target_vbox.set_visible(False)
            self._dialog.set_title(_("Set Language"))

    def _on_current_radio_toggled(self, radio_button):
        """Save the selected target."""
        gaupol.conf.spell_check.target = (
            gaupol.targets.CURRENT
            if radio_button.get_active()
            else gaupol.targets.ALL)

    def _on_main_radio_toggled(self, radio_button):
        """Save the selected field."""
        gaupol.conf.spell_check.field = (
            gaupol.fields.MAIN_TEXT
            if radio_button.get_active()
            else gaupol.fields.TRAN_TEXT)

    def _on_language_list_row_activated(self, list_box, row):
        """Select the language of the activated row."""
        row.gaupol_radio.set_active(True)

    def _on_language_radio_toggled(self, radio):
        """Save the selected language."""
        if radio.get_active():
            gaupol.conf.spell_check.language = radio.gaupol_locale



# -*- coding: utf-8 -*-

# Copyright (C) 2013 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Toast notification that can be overlaid on another widget."""

import gaupol

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

class Toast(Gtk.Box):

    """Toast notification that can be overlaid on another widget."""

    def __init__(self):
        """Initialize a :class:`Toast` instance."""
        GObject.GObject.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        self._hide_id = None
        self._label = Gtk.Label()
        self.add_css_class("gaupol-toast")
        self.set_halign(Gtk.Align.CENTER)
        self.set_valign(Gtk.Align.END)
        self._init_widgets()

    def flash_text(self, text, duration=3):
        """Show toast with `text` for `duration` seconds."""
        if not text:
            return self._hide()
        self.set_text(text)
        self._hide_id = gaupol.util.delay_add(duration * 1000, self._hide)

    def get_text(self):
        """Return text shown in the toast."""
        return self._label.get_text()

    def _hide(self, *args):
        """Hide the toast."""
        self.set_visible(False)
        self._hide_id = None

    def _init_widgets(self):
        """Initialize widgets contained in the toast."""
        gaupol.util.pack_start(self, self._label)
        button = Gtk.Button.new_from_icon_name("window-close-symbolic")
        button.add_css_class("flat")
        button.add_css_class("circular")
        button.connect("clicked", self._hide)
        gaupol.util.pack_start(self, button)

    def set_text(self, text):
        """Show `text` in the toast."""
        if not text:
            return self._hide()
        if self._hide_id is not None:
            GLib.source_remove(self._hide_id)
        self._label.set_text(text)
        self.set_visible(True)



# -*- coding: utf-8 -*-

# Copyright (C) 2016 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""CSS styles and helper functions for styling."""

import gaupol

from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

CSS_FILE = gaupol.DATA_DIR / "gaupol.css"
CSS = CSS_FILE.read_text()

css_provider = None

def _get_editor_font_css():
    """Return CSS for custom editor font."""
    font_desc = Pango.FontDescription("monospace")
    if (gaupol.conf.editor.custom_font and
        gaupol.conf.editor.use_custom_font):
        font_desc = Pango.FontDescription(gaupol.conf.editor.custom_font)
    css = """
    .gaupol-custom-font {{
        font-family: {family},monospace;
        font-size: {size}pt;
        font-weight: {weight};
    }}""".format(
        family=font_desc.get_family().split(",")[0],
        size=int(round(font_desc.get_size() / Pango.SCALE)),
        weight=int(font_desc.get_weight()))
    css = css.replace("font-size: 0pt;", "")
    css = css.replace("font-weight: 0;", "")
    css = "\n".join(filter(lambda x: x.strip(), css.splitlines()))
    return css

def load_css(widget):
    """Load CSS rules from file and conf for `widget`."""
    global css_provider
    if css_provider is None:
        css_provider = Gtk.CssProvider()
        priority = Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        Gtk.StyleContext.add_provider_for_display(
            Gdk.Display.get_default(), css_provider, priority)
    _update_css()

def _update_css(*args, **kwargs):
    """Reload CSS rules from file and conf into the shared provider."""
    if css_provider is None: return
    css = "\n".join((CSS, _get_editor_font_css()))
    # load_from_string is new in GTK-4.12; load_from_data
    # is deprecated since 4.12 and its signature changed.
    css_provider.load_from_string(css)

def use_font(widget, font):
    """Use `font` ("custom" or "monospace") for `widget`."""
    if not font: return
    load_css(widget)
    widget.add_css_class({
        "custom": "gaupol-custom-font",
        "monospace": "monospace",
    }[font])

gaupol.conf.editor.connect("notify::custom_font", _update_css)
gaupol.conf.editor.connect("notify::use_custom_font", _update_css)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Searching for and replacing text."""

import aeidon
import gaupol

class SearchAgent(aeidon.Delegate):

    """Searching for and replacing text."""

    def __init__(self, master):
        """Initialize a :class:`SearchAgent` instance."""
        aeidon.Delegate.__init__(self, master)
        self._search_dialog = None

    def _on_search_dialog_response(self, *args):
        """Hide the search dialog."""
        self._search_dialog.set_visible(False)

    @aeidon.deco.export
    def _on_find_and_replace_activate(self, *args):
        """Search for and replace text."""
        if self._search_dialog is not None:
            return self._search_dialog.present()
        self._search_dialog = gaupol.SearchDialog(self.window, self)
        aeidon.util.connect(self, "_search_dialog", "response")
        # Do not destroy the dialog, but rather hide based on response.
        self._search_dialog.set_hide_on_close(True)
        self._search_dialog.present()

    @aeidon.deco.export
    def _on_find_next_activate(self, *args):
        """Search forwards for same text."""
        self._search_dialog.next()

    @aeidon.deco.export
    def _on_find_previous_activate(self, *args):
        """Search backwards for same text."""
        self._search_dialog.previous()



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Cell renderer for multiline text data."""

import aeidon
import difflib
import gaupol
import re
import traceback

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

class CellTextView(Gtk.TextView, Gtk.CellEditable):

    """A :class:`Gtk.TextView` suitable for cell renderer use."""

    __gproperties__ = {
        "editing-canceled": (bool,
                             "Editing canceled",
                             "Editing canceled",
                             False,
                             GObject.ParamFlags.READWRITE),

    }
    __gtype_name__ = "CellTextView"

    def __init__(self, text_buffer=None):
        """Initialize a :class:`CellTextView` instance."""
        GObject.GObject.__init__(self)
        gaupol.util.prepare_text_view(self)
        controller = Gtk.EventControllerKey()
        controller.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        controller.connect("key-pressed", self._on_key_pressed)
        self.add_controller(controller)

    def do_editing_done(self, *args):
        """End editing."""
        pass

    def do_remove_widget(self, *args):
        """Remove widget."""
        pass

    def do_start_editing(self, *args):
        """Start editing."""
        # Don't let anyone else handle clicks
        # that happen within the text view.
        gesture = Gtk.GestureClick()
        gesture.set_button(0)
        gesture.connect("pressed", lambda gesture, *args: (
            gesture.set_state(Gtk.EventSequenceState.CLAIMED)))
        self.add_controller(gesture)

    def get_text(self):
        """Return text."""
        text_buffer = self.get_buffer()
        start, end = text_buffer.get_bounds()
        return text_buffer.get_text(start, end, False)

    def _on_key_pressed(self, controller, keyval, keycode, state):
        """Toggle italicization if Control+I pressed."""
        if (state & Gdk.ModifierType.CONTROL_MASK and
            keyval in (Gdk.KEY_i, Gdk.KEY_I)):
            self._on_toggle_italic()
            return True
        return False

    def _on_toggle_italic(self, *args):
        """Add or remove italic tags around selection."""
        if not getattr(gaupol, "italic_tag", None): return
        if not getattr(gaupol, "italicize", None): return
        text_buffer = self.get_buffer()
        start, end = text_buffer.get_bounds()
        text = text_buffer.get_text(start, end, False)
        bounds = text_buffer.get_selection_bounds()
        if not bounds:
            # If no selection, replicate cursor position.
            cursor = text_buffer.get_insert()
            cursor = text_buffer.get_iter_at_mark(cursor)
            bounds = (cursor, cursor)
        a, b = [x.get_offset() for x in bounds]
        # Strip possible existing italic tags around selection.
        # If none found, then selection should be italicized.
        text1 = re.sub(gaupol.italic_tag.pattern + r"\Z", "", text[:a])
        text2 = re.sub(gaupol.italic_tag.pattern, "", text[a:b])
        text3 = re.sub(r"\A" + gaupol.italic_tag.pattern, "", text[b:])
        if text1 + text2 + text3 == text:
            text2 = gaupol.italicize(text2)
            match = gaupol.italic_tag.match(text2)
            a += match.end() - match.start()
            b += match.end() - match.start()
        else:
            a = len(text1)
            b = len(text1) + len(text2)
        text_buffer.set_text(text1 + text2 + text3)
        ins = text_buffer.get_iter_at_offset(a)
        bound = text_buffer.get_iter_at_offset(b)
        text_buffer.select_range(ins, bound)

    def set_text(self, text):
        """Set text."""
        self.get_buffer().set_text(text)

class MultilineCellRenderer(Gtk.CellRendererText):

    """
    Cell renderer for multiline text data.

    If :attr:`gaupol.conf.editor.show_lengths_cell` is ``True``, line lengths
    are shown at the end of each line.
    """

    __gtype_name__ = "MultilineCellRenderer"

    def __init__(self):
        """Initialize a :class:`MultilineCellRenderer` instance."""
        GObject.GObject.__init__(self)
        self._show_lengths = gaupol.conf.editor.show_lengths_cell
        self._text = ""
        gaupol.conf.connect_notify("editor", "show_lengths_cell", self)
        aeidon.util.connect(self, self, "notify::text")

    def do_start_editing(self, event, widget, path, bg_area, cell_area, flags):
        """Initialize and return a :class:`CellTextView` widget."""
        editor = CellTextView()
        gaupol.style.use_font(editor, "custom")
        editor.set_text(self._text)
        editor.set_size_request(cell_area.width, cell_area.height)
        editor.set_left_margin(self.props.xpad)
        editor.set_right_margin(self.props.xpad)
        editor.set_top_margin(self.props.ypad)
        editor.set_bottom_margin(self.props.ypad)
        editor.gaupol_path = path
        controller = Gtk.EventControllerFocus()
        controller.connect("leave", self._on_editor_focus_leave, editor)
        editor.add_controller(controller)
        controller = Gtk.EventControllerKey()
        controller.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        controller.connect("key-pressed", self._on_editor_key_pressed, editor)
        editor.add_controller(controller)
        return editor

    def _on_conf_editor_notify_show_lengths_cell(self, *args):
        """Hide or show line lengths if ``conf`` changed."""
        self._show_lengths = gaupol.conf.editor.show_lengths_cell

    def _on_editor_focus_leave(self, controller, editor):
        """End editing."""
        # The leave event is delivered in the middle of the focus
        # change and removing the focused editor there sends GTK's
        # focus handling into an infinite loop. Defer to idle and
        # skip if editing already ended (editor unparented).
        def end_editing():
            if editor.get_parent() is None: return
            editor.remove_widget()
            self.emit("editing-canceled")
        gaupol.util.idle_add(end_editing)

    def _on_editor_key_pressed(self, controller, keyval, keycode, state, editor):
        """End editing if ``Enter`` or ``Escape`` pressed."""
        if (state &
            (Gdk.ModifierType.SHIFT_MASK |
             Gdk.ModifierType.CONTROL_MASK)): return False
        if keyval in (Gdk.KEY_Return, Gdk.KEY_KP_Enter):
            editor.editing_done()
            editor.remove_widget()
            self.emit("edited", editor.gaupol_path, editor.get_text())
            return True
        if keyval == Gdk.KEY_Escape:
            editor.editing_done()
            editor.remove_widget()
            self.emit("editing-canceled")
            return True
        return False

    def _on_notify_text(self, *args):
        """Set markup by adding line lengths to text."""
        # Since GTK-3.6, the notify::text signal seems to get
        # emitted insanely often even if text hasn't changed at
        # all. Let's try to keep this callback as fast as possible.
        self._text = self.props.text
        self.props.markup = self._text_to_markup(self.props.text,
                                                 self._show_lengths,
                                                 gaupol.conf.editor.length_unit)

    def set_show_lengths(self, show_lengths):
        """Show or hide line lengths, overriding ``conf``."""
        self._show_lengths = show_lengths
        gaupol.conf.disconnect_notify("editor", "show_lengths_cell", self)

    @aeidon.deco.memoize(1000)
    def _text_to_markup(self, text, show_lengths, length_unit):
        """Return `text` rendered as markup for display."""
        # We don't actually use the length_unit argument,
        # but do need it accounted for in memoized values.
        if not text: return ""
        markup = GLib.markup_escape_text(text)
        if not show_lengths: return markup
        lines = markup.split("\n")
        lengths = gaupol.ruler.get_lengths(text)
        return "\n".join((f"{lines[i]} <small>[{lengths[i]:d}]</small>"
                          if lines[i] else lines[i]
                          for i in range(len(lines))))

class MultilineDiffCellRenderer(MultilineCellRenderer):

    """Cell renderer for multiline diffed text data."""

    __gtype_name__ = "MultilineDiffCellRenderer"

    ref_text = GObject.Property(type=str, default="")
    ref_type = GObject.Property(type=int, default=-1)

    COLOR_CHANGE = gaupol.conf.general.diff_color_change
    COLOR_DELETE = gaupol.conf.general.diff_color_delete
    COLOR_INSERT = gaupol.conf.general.diff_color_insert

    def __init__(self):
        """Initialize a :class:`MultilineDiffCellRenderer` instance."""
        MultilineCellRenderer.__init__(self)
        self._ref_text = ""

    def _add_diff_markup(self, a, b):
        """Return `a` with markup for parts different from `b`."""
        matcher = difflib.SequenceMatcher(a=a, b=b)
        ops = matcher.get_opcodes()
        return "".join(self._highlight(a[i1:i2], tag)
                       for tag, i1, i2, j1, j2 in ops)

    def _get_diff_color(self, tag):
        """Return color to use to highlight changes of type `tag`."""
        if tag == "insert" and self.props.ref_type > 0: return self.COLOR_INSERT
        if tag == "insert" and self.props.ref_type < 0: return self.COLOR_DELETE
        if tag == "delete" and self.props.ref_type > 0: return self.COLOR_DELETE
        if tag == "delete" and self.props.ref_type < 0: return self.COLOR_INSERT
        return self.COLOR_CHANGE

    def _highlight(self, text, tag):
        """Return `text` as markup highlighting changes of type `tag`."""
        if tag == "equal": return text
        color = self._get_diff_color(tag)
        return f'<span background="{color}">{text}</span>'

    def _on_notify_text(self, *args):
        """Set markup by adding line lengths to text."""
        # Since GTK-3.6, the notify::text signal seems to get
        # emitted insanely often even if text hasn't changed at
        # all. Let's try to keep this callback as fast as possible.
        self._text = self.props.text
        self._ref_text = self.props.ref_text
        self.props.markup = self._text_to_markup(self.props.text,
                                                 self.props.ref_text,
                                                 self._show_lengths,
                                                 gaupol.conf.editor.length_unit)

    @aeidon.deco.memoize(1000)
    def _text_to_markup(self, text, ref_text, show_lengths, length_unit):
        """Return `text` rendered as markup for display."""
        # We don't actually use the length_unit argument,
        # but do need it accounted for in memoized values.
        if not text: return ""
        text = GLib.markup_escape_text(text)
        ref_text = GLib.markup_escape_text(ref_text)
        try:
            text = self._add_diff_markup(text, ref_text)
        except Exception:
            traceback.print_exc()
        if not show_lengths: return text
        lines = text.split("\n")
        lengths = gaupol.ruler.get_lengths(text)
        return "\n".join((f"{lines[i]} <small>[{lengths[i]:d}]</small>"
                          if lines[i] else lines[i]
                          for i in range(len(lines))))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Paths to files and directories used."""

import os
import sys

from pathlib import Path

def get_config_home_directory():
    """Return path to the user's configuration directory."""
    if sys.platform == "win32":
        directory = os.environ.get("APPDATA") or Path.home()
        return (Path(directory) / "Gaupol").resolve()
    directory = os.environ.get("XDG_CONFIG_HOME") or Path.home() / ".config"
    return (Path(directory) / "gaupol").resolve()

def get_data_directory():
    """Return path to the aeidon data directory."""
    # Data files are shipped inside the Python package,
    # which works the same in all contexts.
    return Path(__file__).resolve().parent / "data"

def get_data_home_directory():
    """Return path to the user's data directory."""
    if sys.platform == "win32":
        directory = os.environ.get("APPDATA") or Path.home()
        return (Path(directory) / "Gaupol").resolve()
    default = Path.home() / ".local" / "share"
    directory = os.environ.get("XDG_DATA_HOME") or default
    return (Path(directory) / "gaupol").resolve()

CONFIG_HOME_DIR = get_config_home_directory()
DATA_DIR = get_data_directory()
DATA_HOME_DIR = get_data_home_directory()



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Metadata store for one item in a desktop-style file."""

from aeidon.i18n import _

class MetadataItem:

    """
    Metadata store for one item in a desktop-style file.

    :ivar fields: Dictionary mapping field names to their string values

    Common translated fields with custom handling are ``Name`` and
    ``Description``; arbitrary fields are accessible with :meth:`get_field`.
    Strings ``True`` and ``False`` are used for boolean fields.

    For the string syntax, see freedesktop.org_'s Desktop Entry
    Specification_, of which translations are the notable exception:
    fields are stored in English only and translated at runtime.

    .. _freedesktop.org: https://www.freedesktop.org/
    .. _Specification: https://www.freedesktop.org/wiki/Specifications/desktop-entry-spec/
    """

    def __init__(self, fields=None):
        """Initialize a :class:`MetadataItem` instance."""
        self.fields = fields or {}

    def get_description(self, localize=True):
        """Return description as defined by the ``Description`` field."""
        description = self.get_field("Description")
        return _(description) if localize else description

    def get_field(self, name, fallback=None):
        """Return the string value of field or `fallback`."""
        if not name in self.fields:
            return fallback
        return self.fields[name]

    def get_field_boolean(self, name, fallback=None):
        """Return the boolean value of field or `fallback`."""
        if not name in self.fields:
            return fallback
        value = self.fields[name]
        if value == "True":
            return True
        if value == "False":
            return False
        raise ValueError(f"Invalid boolean value: {value!r}")

    def get_field_list(self, name, fallback=None):
        """Return the list of strings value of field or `fallback`."""
        if not name in self.fields:
            return fallback
        lst = self.fields[name].split(";")
        if not lst[-1]: lst.pop(-1)
        return lst

    def get_name(self, localize=True):
        """Return name as defined by the ``Name`` field."""
        name = self.get_field("Name")
        return _(name) if localize else name

    def has_field(self, name):
        """Return ``True`` if field exists."""
        return (name in self.fields)

    def set_field(self, name, value):
        """Set the string value of field."""
        self.fields[name] = str(value)



# -*- coding: utf-8 -*-

# Copyright (C) 2006 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Baseclass and wrapper for :class:`Gtk.Builder` constructed dialogs."""

import gaupol

from gi.repository import Gtk
from pathlib import Path

class BuilderDialog:

    """
    Baseclass and wrapper for :class:`Gtk.Builder` constructed dialogs.

    :cvar _widgets: List of names of widgets to be assigned as attributes

    All widgets defined in :attr:`_widgets` are assigned as instance variables
    with names preceded by a single underscore. All signals defined in the UI
    definition file are connected to ``self``. All :func:`getattr` calls not
    found in ``self`` are delegated to :attr:`self._dialog` allowing ``self``
    to look and act like a :class:`Gtk.Dialog`.
    """

    _widgets = []

    def __init__(self, ui_file_path):
        """Initialize a :class:`BuilderDialog` instance from `ui_file_path`."""
        ui_file_path = Path(ui_file_path)
        if not ui_file_path.is_absolute():
            ui_file_path = gaupol.DATA_DIR / "ui" / ui_file_path
        # Signal handlers are resolved against self already when
        # the UI definition file is parsed, i.e. before _dialog is set.
        self._builder = Gtk.Builder(self)
        self._builder.set_translation_domain("gaupol")
        self._builder.add_from_file(str(ui_file_path))
        self._dialog = self._builder.get_object("dialog")
        self._set_attributes(self._widgets)

    def __getattr__(self, name):
        """Return attribute from :attr:`_dialog`."""
        if name == "_dialog":
            raise AttributeError(name)
        return getattr(self._dialog, name)

    def _set_attributes(self, widgets, prefix=None):
        """Assign all names in `widgets` as attributes of `self`."""
        for name in widgets:
            widget = self._builder.get_object(name)
            if prefix and name.startswith(prefix):
                name = name.replace(prefix, "")
            setattr(self, f"_{name}", widget)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Text markup for the SubRip format."""

import aeidon
import re

class SubRip(aeidon.Markup):

    """
    Text markup for the SubRip format.

    SubRip format is assumed (based on the SubRip application GUI) to contain
    the following HTML-style tags, in either lower- or upper case.

     * ``<b>.........................</b>``
     * ``<i>.........................</i>``
     * ``<u>.........................</u>``
     * ``<font color="#RRGGBB">...</font>``
    """

    _flags = re.DOTALL | re.MULTILINE | re.IGNORECASE
    format = aeidon.formats.SUBRIP

    def bolden(self, text, bounds=None):
        """Return bolded `text`."""
        a, z = bounds or (0, len(text))
        return "".join((text[:a], f"<b>{text[a:z]}</b>", text[z:]))

    def clean(self, text):
        """Return `text` with less ugly markup."""
        # Remove tags that are immediately closed after opening.
        text = self._substitute(text, r"<([a-z]+)[^<]*?>( *)</\1>", r"\2")
        # Remove tags that are immediately opened after closing.
        # Don't apply to tags that have values, such as 'font',
        # because then we'd need to check the value is the same too.
        text = self._substitute(text, r"</([a-z]+)>( *)<\1>", r"\2")
        # Remove or relocate space right after an opening tag.
        text = self._substitute(text, r" ?(<(?!/)[^>]+?>) ", r" \1")
        # Remove or relocate space right before a closing tag.
        text = self._substitute(text, r" (</[^>]+?>) ?", r"\1 ")
        return text

    def colorize(self, text, color, bounds=None):
        """Return `text` colorized to hexadecimal value."""
        a, z = bounds or (0, len(text))
        target = f'<font color="#{color}">{text[a:z]}</font>'
        return "".join((text[:a], target, text[z:]))

    @property
    def italic_tag(self):
        """Regular expression for an italic markup tag."""
        return self._get_regex(r"</?i>")

    def italicize(self, text, bounds=None):
        """Return italicized `text`."""
        a, z = bounds or (0, len(text))
        return "".join((text[:a], f"<i>{text[a:z]}</i>", text[z:]))

    def _main_decode(self, text):
        """Return `text` with decodable markup decoded."""
        text = self._decode_b(text, r"<b>(.*?)</b>", 1)
        text = self._decode_i(text, r"<i>(.*?)</i>", 1)
        text = self._decode_u(text, r"<u>(.*?)</u>", 1)
        pattern = r'<font color="#([0-9a-fA-F]{6})">(.*?)</font>'
        return self._decode_c(text, pattern, 1, 2)

    @property
    def tag(self):
        """Regular expression for any markup tag."""
        return self._get_regex(r"<.*?>")

    def underline(self, text, bounds=None):
        """Return underlined `text`."""
        a, z = bounds or (0, len(text))
        return "".join((text[:a], f"<u>{text[a:z]}</u>", text[z:]))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Building and updating dynamic menus."""

import aeidon
import contextlib
import gaupol

from aeidon.i18n   import _
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk

class MenuAgent(aeidon.Delegate):

    """Building and updating dynamic menus."""

    def __init__(self, master):
        """Initialize a :class:`MenuAgent` instance."""
        aeidon.Delegate.__init__(self, master)
        self._columns_popup = None
        self._view_popup = None
        self._init_signal_handlers()

    def _get_recent_paths(self, limit=10):
        """Return a list of recently used subtitle file paths."""
        mime_types = [x.mime_type for x in aeidon.formats]
        items = [x for x in self.recent_manager.get_items()
                 if x.has_application("gaupol")
                 and x.get_mime_type() in mime_types]
        if not gaupol.conf.recent.show_not_found:
            # Can be really slow with network drives.
            # https://github.com/otsaloma/gaupol/issues/175
            items = [x for x in items if x.exists()]
        items.sort(key=lambda x: x.get_modified().to_unix(), reverse=True)
        return [aeidon.util.uri_to_path(x.get_uri()) for x in items[:limit]]

    def _init_signal_handlers(self):
        """Initialize signal handlers."""
        self.connect("page-added", self._update_projects_menu)
        self.connect("page-closed", self._update_projects_menu)
        self.connect("page-saved", self._update_projects_menu)
        self.connect("page-switched", self._update_projects_menu)
        self.connect("pages-reordered", self._update_projects_menu)
        self.connect("init-done", self._update_recent_menus)
        self.connect("page-added", self._update_recent_menus)
        self.connect("page-saved", self._update_recent_menus)

    @aeidon.deco.export
    def _on_activate_project_activate(self, action, parameter):
        """Activate the requested page."""
        index = int(parameter.get_string())
        self.set_current_page(self.pages[index])

    @aeidon.deco.export
    def _on_open_recent_main_file_activate(self, action, *args):
        """Open recent file as main document."""
        if not action.gaupol_path.is_file():
            return self.flash_message(_("File not found"))
        self.open_main(action.gaupol_path)

    @aeidon.deco.export
    def _on_open_recent_translation_file_activate(self, action, *args):
        """Open recent file as translation document."""
        if not action.gaupol_path.is_file():
            return self.flash_message(_("File not found"))
        self.open_translation(action.gaupol_path)

    @aeidon.deco.export
    def _on_view_pressed(self, gesture, n_press, x, y):
        """Display a right-click pop-up menu to edit data."""
        view = gesture.get_widget()
        bx, by = view.convert_widget_to_bin_window_coords(int(x), int(y))
        value = view.get_path_at_pos(bx, by)
        if value is None: return
        path, column = value[:2]
        row = gaupol.util.tree_path_to_row(path)
        if not row in view.get_selected_rows():
            view.set_cursor(path, column)
            view.update_headers()
        if self._view_popup is None:
            path = gaupol.DATA_DIR / "ui" / "view-popup.ui"
            builder = Gtk.Builder.new_from_file(str(path))
            self._view_popup = builder.get_object("view-popup")
        self._show_popover_menu(gesture, x, y, self._view_popup)

    @aeidon.deco.export
    def _on_view_header_pressed(self, gesture, n_press, x, y):
        """Display a column visibility pop-up menu."""
        if self._columns_popup is None:
            path = gaupol.DATA_DIR / "ui" / "columns-popup.ui"
            builder = Gtk.Builder.new_from_file(str(path))
            self._columns_popup = builder.get_object("columns-popup")
        self._show_popover_menu(gesture, x, y, self._columns_popup)

    def _show_popover_menu(self, gesture, x, y, model):
        """Show a pop-up menu for `model` at coordinates."""
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        menu = Gtk.PopoverMenu.new_from_model(model)
        menu.set_parent(gesture.get_widget())
        rect = Gdk.Rectangle()
        rect.x, rect.y = int(x), int(y)
        rect.width = rect.height = 1
        menu.set_pointing_to(rect)
        menu.set_has_arrow(False)
        menu.connect("closed", lambda menu: GLib.idle_add(menu.unparent))
        menu.popup()

    def _update_projects_menu(self, *args):
        """Update the project menu list of projects."""
        menu = self.get_menubar_section("projects-placeholder")
        # Menubar not available when running unit tests.
        if menu is None: return
        menu.remove_all()
        current = self.get_current_page()
        for i, page in enumerate(self.pages):
            label = page.get_main_basename()
            if len(label) > 100:
                label = label[:100] + "…"
            action = f"win.activate-project::{i:d}"
            menu.append(label, action)
            if page is current:
                action = self.get_action("activate-project")
                action.set_state(str(i))

    def _update_recent_main_menu(self, *args):
        """Update the file menu list of recent main files."""
        menu = self.get_menubar_section("open-recent-main-placeholder")
        # Menubar not available when running unit tests.
        if menu is None: return
        menu.remove_all()
        for i, path in enumerate(self._get_recent_paths()):
            label = path.name
            if len(label) > 100:
                label = label[:100] + "…"
            action = f"win.open-recent-main-file-{i:d}"
            menu.append(label, action)
            action = action.replace("win.", "")
            if self.get_action(action):
                # If action i exists, update the file path.
                self.get_action(action).gaupol_path = path
            else:
                # Otherwise, create the action and add to self.window.
                # XXX: For some reason, this can sometimes fail.
                with contextlib.suppress(Exception):
                    ao = gaupol.Action(action)
                    ao.gaupol_path = path
                    callback = self._on_open_recent_main_file_activate
                    ao.connect("activate", callback)
                    self.window.add_action(ao)

    def _update_recent_menus(self, *args):
        """Update the file menu lists of recent files."""
        self._update_recent_main_menu()
        self._update_recent_translation_menu()
        # Update enabled states of added actions.
        self.update_gui()

    def _update_recent_translation_menu(self, *args):
        """Update the file menu list of recent translation files."""
        menu = self.get_menubar_section("open-recent-translation-placeholder")
        # Menubar not available when running unit tests.
        if menu is None: return
        menu.remove_all()
        for i, path in enumerate(self._get_recent_paths()):
            label = path.name
            if len(label) > 100:
                label = label[:100] + "…"
            action = f"win.open-recent-translation-file-{i:d}"
            menu.append(label, action)
            action = action.replace("win.", "")
            if not self.get_action(action):
                ao = gaupol.OpenRecentTranslationFileAction(action)
                callback = self._on_open_recent_translation_file_activate
                ao.connect("activate", callback)
                self.window.add_action(ao)
            self.get_action(action).gaupol_path = path



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Reading, writing and manipulating text-based subtitle files.

:mod:`aeidon` is a Python package that provides classes and functions for
dealing with text-based subtitle files of many different formats. Functions
exist for reading and writing subtitle files as well as manipulating subtitle
data, i.e. positions (times or frames) and texts. Two examples of basic use
of the :mod:`aeidon` package follow below.

Converting a file from the SubRip format to the MicroDVD format::

    project = aeidon.Project()
    project.open_main("/path/to/file.srt", "utf_8")
    project.set_framerate(aeidon.framerates.FPS_23_976)
    project.save_main(aeidon.files.new(aeidon.formats.MICRODVD,
                                       "/path/to/file.sub",
                                       "utf_8"))

Making all subtitles in a file appear two seconds earlier::

    project = aeidon.Project()
    project.open_main("/path/to/file.srt", "utf_8")
    project.shift_positions(None, aeidon.as_seconds(-2))
    project.save_main()

:mod:`aeidon` handles positions as either times (as strings of form
``HH:MM:SS.SSS``), frames (as integers) or in some cases seconds (as floats).
All these can be used with functions that edit subtitle data regardless of what
the native position type of the file format used is.

:mod:`aeidon` handles two separate documents that comprise a project -- a main
and a translation document. These correspond to separate files, but the
subtitles are common since positions are shared.

:mod:`aeidon` includes an undo/redo-system. Any subtitle data-editing methods
of :class:`aeidon.Project` (ones marked with the :func:`aeidon.deco.revertable`
decorator) can be undone and redone. If using :mod:`aeidon` in a context where
reverting actions is never needed, greater flexibility can be achieved by
accessing the subtitles directly (via :attr:`aeidon.Project.subtitles`).

:var CONFIG_HOME_DIR: Path to the user's local configuration directory
:var DATA_DIR: Path to the aeidon data directory
:var DATA_HOME_DIR: Path to the user's local data directory
:var RE_ANY_TAG: Regular expression for markup tags of any format

:var align_methods: Enumerations for subtitle align methods
:var documents: Enumerations for document types
:var formats: Enumerations for subtitle file format types
:var framerates: Enumerations for framerate types
:var modes: Enumerations for position unit types
:var newlines: Enumerations for newline character types
:var players: Enumerations for video player application types
:var registers: Enumerations for action action reversion register types
"""

import re
import sys

__version__ = "2.0"

RUNNING_SPHINX = (sys.argv[0].endswith("autogen.py") or
                  sys.argv[0].endswith("sphinx-build"))

RE_ANY_TAG = re.compile(r"(^[/\\_]+|<.*?>|\{.*?\})")

try:
    import gi
    gi.require_version("Spelling", "1")
except Exception:
    pass

from aeidon.paths import CONFIG_HOME_DIR
from aeidon.paths import DATA_DIR
from aeidon.paths import DATA_HOME_DIR
from aeidon.position import as_frame
from aeidon.position import as_seconds
from aeidon.position import as_time
from aeidon.position import is_frame
from aeidon.position import is_seconds
from aeidon.position import is_time
from aeidon import deco
from aeidon import i18n
from aeidon import util
from aeidon import temp
from aeidon.delegate import Delegate
from aeidon.singleton import Singleton
from aeidon.mutables import ObservableDict
from aeidon.mutables import ObservableList
from aeidon.mutables import ObservableSet
from aeidon.observable import Observable
from aeidon.errors import Error
from aeidon.errors import AffirmationError
from aeidon.errors import FormatError
from aeidon.errors import ParseError
from aeidon.errors import ProcessError
from aeidon.enum import EnumerationItem
from aeidon.enum import Enumeration
from aeidon.enums import align_methods
from aeidon.enums import documents
from aeidon.enums import formats
from aeidon.enums import framerates
from aeidon.enums import modes
from aeidon.enums import newlines
from aeidon.enums import players
from aeidon.enums import registers
from aeidon import encodings
from aeidon import languages
from aeidon import countries
from aeidon import locales
from aeidon import scripts
from aeidon.metadata import MetadataItem
from aeidon.calculator import Calculator
from aeidon.finder import Finder
from aeidon.parser import Parser
from aeidon.liner import Liner
from aeidon import containers
from aeidon.subtitle import Subtitle
from aeidon.file import SubtitleFile
from aeidon import files
from aeidon.markup import Markup
from aeidon import markups
from aeidon.markupconv import MarkupConverter
from aeidon.pattern import Pattern
from aeidon.patternman import PatternManager
from aeidon.clipboard import Clipboard
from aeidon.revertable import RevertableAction
from aeidon.revertable import RevertableActionGroup
from aeidon.spell import SpellChecker
from aeidon.spell import SpellCheckNavigator
from aeidon.spell import SpellCheckTokenizer
from aeidon import agents
from aeidon.project import Project
from aeidon.unittest import TestCase



# -*- coding: utf-8 -*-

# Copyright (C) 2019 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Checking and correcting spelling."""

import aeidon
import contextlib

with contextlib.suppress(Exception):
    from gi.repository import GtkSource
    from gi.repository import Spelling

class SpellChecker(aeidon.SpellChecker):

    """Checking the spelling of an individual word."""

    def attach(self, text_view):
        """Attach inline spell-check to `text_view`."""
        # Spelling.TextBufferAdapter requires a GtkSource.Buffer,
        # replace the text view's blank default buffer with one.
        text_buffer = GtkSource.Buffer()
        text_buffer.set_highlight_matching_brackets(False)
        text_view.set_buffer(text_buffer)
        adapter = Spelling.TextBufferAdapter.new(text_buffer, self.checker)
        text_view.set_extra_menu(adapter.get_menu_model())
        text_view.insert_action_group("spelling", adapter)
        adapter.set_enabled(True)
        return adapter



# -*- coding: utf-8 -*-

# Copyright (C) 2006 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Functions to calculate line lengths and to show them in widgets."""

import aeidon
import gaupol

from gi.repository import GObject
from gi.repository import Graphene
from gi.repository import Gtk
from gi.repository import Pango

class _Ruler:

    """Measurer of line lengths in various units."""

    def __init__(self):
        """Initialize a :class:`_Ruler` instance."""
        self._em_length = None
        self._label = Gtk.Label()
        self._update_em_length()

    def get_char_length(self, text, strip=False, floor=False):
        """Return length of `text` measured in characters."""
        text = aeidon.RE_ANY_TAG.sub("", text) if strip else text
        return len(text.replace("\n", " "))

    def get_em_length(self, text, strip=False, floor=False):
        """Return length of `text` measured in ems."""
        text = aeidon.RE_ANY_TAG.sub("", text) if strip else text
        self._label.set_text(text.replace("\n", " "))
        width = self._label.measure(Gtk.Orientation.HORIZONTAL, -1).natural
        length = width / self._em_length
        return int(length) if floor else length

    def _update_em_length(self):
        """Update the length of em based on font rendering."""
        text = "abcdefghijklmnopqrstuvwxyz"
        self._label.set_text(text)
        width = self._label.measure(Gtk.Orientation.HORIZONTAL, -1).natural
        # About 0.55 em per a-z average character.
        # https://bugzilla.gnome.org/show_bug.cgi?id=763589
        self._em_length = width / (0.55 * len(text))

_ruler = _Ruler()

class _Margin(Gtk.Widget):

    """Widget that shows line lengths in a text view's right gutter."""

    def __init__(self, text_view):
        """Initialize a :class:`_Margin` instance."""
        GObject.GObject.__init__(self)
        self._handlers = []
        self._text_view = text_view
        self._vadjustment = None
        text_buffer = text_view.get_buffer()
        handler_id = text_buffer.connect("changed", self._on_buffer_changed)
        self._handlers.append((text_buffer, handler_id))
        handler_id = text_view.connect(
            "notify::vadjustment", self._on_notify_vadjustment)
        self._handlers.append((text_view, handler_id))
        self._on_notify_vadjustment()
        self._update_width()

    def disconnect_all(self):
        """Disconnect all signal handlers connected to other objects."""
        for obj, handler_id in self._handlers:
            obj.disconnect(handler_id)
        self._handlers = []
        if self._vadjustment is not None:
            self._vadjustment.disconnect(self._vadjustment_handler)
            self._vadjustment = None

    def do_snapshot(self, snapshot):
        """Draw line lengths aligned with the text view's lines."""
        text_view = self._text_view
        text_buffer = text_view.get_buffer()
        start, end = text_buffer.get_bounds()
        text = text_buffer.get_text(start, end, False)
        if not text: return
        lengths = get_lengths(text)
        color = self.get_color()
        layout = Pango.Layout(text_view.get_pango_context())
        width = self.get_width()
        for i in range(len(lengths)):
            ok, itr = text_buffer.get_iter_at_line(i)
            if not ok: break
            location = text_view.get_iter_location(itr)
            y = text_view.buffer_to_window_coords(
                Gtk.TextWindowType.RIGHT, 0, location.y)[1]
            layout.set_text(str(lengths[i]), -1)
            point = Graphene.Point()
            point.init(width - layout.get_pixel_size()[0] - 4, y)
            snapshot.save()
            snapshot.translate(point)
            snapshot.append_layout(layout, color)
            snapshot.restore()

    def _on_buffer_changed(self, *args):
        """Update width and redraw when text changes."""
        self._update_width()
        self.queue_draw()

    def _on_notify_vadjustment(self, *args):
        """Redraw on scrolling, also if the adjustment is replaced."""
        if self._vadjustment is not None:
            self._vadjustment.disconnect(self._vadjustment_handler)
        self._vadjustment = self._text_view.get_vadjustment()
        if self._vadjustment is None: return
        self._vadjustment_handler = self._vadjustment.connect(
            "value-changed", lambda *args: self.queue_draw())

    def _update_width(self):
        """Request enough width for the widest line length."""
        text_buffer = self._text_view.get_buffer()
        start, end = text_buffer.get_bounds()
        text = text_buffer.get_text(start, end, False)
        layout = Pango.Layout(self._text_view.get_pango_context())
        layout.set_text(str(max(get_lengths(text))), -1)
        self.set_size_request(layout.get_pixel_size()[0] + 6, -1)

def connect_text_view(text_view):
    """Connect `text_view` to show line lengths in its margin."""
    if hasattr(text_view, "gaupol_ruler_margin"): return
    margin = _Margin(text_view)
    text_view.set_gutter(Gtk.TextWindowType.RIGHT, margin)
    text_view.gaupol_ruler_margin = margin

def disconnect_text_view(text_view):
    """Disconnect `text_view` from showing line lengths in its margin."""
    if not hasattr(text_view, "gaupol_ruler_margin"): return
    text_view.gaupol_ruler_margin.disconnect_all()
    del text_view.gaupol_ruler_margin
    text_view.set_gutter(Gtk.TextWindowType.RIGHT, None)

def get_length_function(unit):
    """Return a function that returns text length in `unit`."""
    if unit == gaupol.length_units.CHAR:
        return _ruler.get_char_length
    if unit == gaupol.length_units.EM:
        return _ruler.get_em_length
    raise ValueError(f"Invalid length unit: {unit!r}")

def get_lengths(text):
    """Return a sequence of floored line lengths without tags."""
    fun = get_length_function(gaupol.conf.editor.length_unit)
    return [fun(line, strip=True, floor=True) for line in text.split("\n")]
